'''
This module includes classes that should be used with WTForms within Digital Marketplace apps.

They extend the classes found in `wtforms.fields`, and should be used instead of those.
The advantage they provide is that they have been deliberately designed to be used with
the Digital Marketplace frontend toolkit macros. If a particular field is missing, it
should be added to this file.

The main functionality is provided by the mixin class, `DMFieldMixin`. When a derived class
includes `DMFieldMixin` in its base classes then the following extra features are provided:

        - a `hint` property that can be set in the initialiser
        - a `question` property that contains the label text
        - a `value` property for the data that is displayed
        - an `error` property for the field validation error

For more details on how `DMFieldMixin`, see the documentation in `dmutils/forms/mixins.py`.
'''

import datetime
from itertools import chain

import wtforms
import wtforms.fields
from wtforms.utils import unset_value
from wtforms.validators import Length

from .mixins import DMFieldMixin, DMSelectFieldMixin
from .widgets import (
    DMCheckboxInput,
    DMDateInput,
    DMRadioInput,
    DMTextInput,
    DMUnitInput,
)

from .filters import strip_whitespace
from .validators import EmailValidator


__all__ = ['DMBooleanField', 'DMDecimalField', 'DMHiddenField', 'DMIntegerField',
           'DMRadioField', 'DMStringField', 'DMEmailField', 'DMStripWhitespaceStringField',
           'DMDateField']


class DMBooleanField(DMFieldMixin, wtforms.fields.BooleanField):
    widget = DMCheckboxInput(hide_question=True)

    @property
    def href(self):
        # digitalmarketplace-frontend-toolkit suffixes the id with `-1`
        return self._href or super().href + "-1"

    @property
    def options(self):
        return [{"label": self.label.text, "value": self.data}]

    @property
    def value(self):
        if not self.data or self.data == "None":
            return []
        else:
            return [self.data]


class DMDecimalField(DMFieldMixin, wtforms.fields.DecimalField):
    pass


class DMPoundsField(DMDecimalField):
    widget = DMUnitInput()

    unit = "£"
    unit_in_full = "pounds"
    unit_position = "before"
    hint = "For example, 9900.95 for 9900 pounds and 95 pence"


class DMHiddenField(DMFieldMixin, wtforms.fields.HiddenField):
    pass


class DMIntegerField(DMFieldMixin, wtforms.fields.IntegerField):
    pass


class DMRadioField(DMSelectFieldMixin, DMFieldMixin, wtforms.fields.RadioField):
    widget = DMRadioInput()


class DMStringField(DMFieldMixin, wtforms.fields.StringField):
    widget = DMTextInput()


class DMStripWhitespaceStringField(DMFieldMixin, wtforms.fields.StringField):
    widget = DMTextInput()

    def __init__(self, label=None, **kwargs):
        kwargs['filters'] = tuple(chain(
            kwargs.get('filters', ()) or (),
            (
                strip_whitespace,
            ),
        ))
        super().__init__(label, **kwargs)


class DMEmailField(DMStripWhitespaceStringField):
    def __init__(self, label=None, **kwargs):
        kwargs["validators"] = tuple(chain(
            kwargs.pop("validators", ()) or (),
            (
                EmailValidator(),
                Length(max=511, message="Please enter an email address under 512 characters."),
            ),
        ))
        super().__init__(label, **kwargs)


class DMDateField(DMFieldMixin, wtforms.fields.Field):
    '''
    A date field(set) that uses a day, month and year field.

    The data is converted to a `datetime.date`. The year is
    required to be four digits.

    It behaves like a WTForms.FieldForm, but it can be used
    with validators like a normal WTForms.Field.

    >>> from wtforms import Form
    >>> from wtforms.validators import DataRequired
    >>> from werkzeug.datastructures import MultiDict
    >>> formdata = MultiDict({
    ...     'date-day': '31',
    ...     'date-month': '12',
    ...     'date-year': '1999'})
    >>> class DateForm(Form):
    ...     date = DateField(validators=[DataRequired()])
    >>> form = DateForm(formdata)
    >>> form.date.data
    datetime.date(1999, 12, 31)
    '''

    widget = DMDateInput()

    hint = "For example, 31 12 2020"

    # An internal class that defines the fields that make up the DateField.
    #
    # Inheriting from wtforms.FormField has limitations on using validators.
    #
    # Instead, the DateField is composed of a wtforms.FormField that is used
    # to turn the form data into integer values, and we then grab the data.
    #
    # The FormField instance is based on this class.
    class _DateForm(wtforms.Form):
        day = wtforms.fields.IntegerField("Day")
        month = wtforms.fields.IntegerField("Month")
        year = wtforms.fields.IntegerField("Year")

    def __init__(self, label=None, validators=None, hint=None, question_advice=None, separator='-', **kwargs):
        super().__init__(label=label, validators=validators, hint=hint, question_advice=question_advice, **kwargs)
        self.form_field = wtforms.fields.FormField(self._DateForm, separator=separator, **kwargs)

    def _value(self):
        '''
        Return the values that are used to display the form

        Overrides wtforms.Field._value().
        '''
        if self.raw_data:
            return self.raw_data[0]
        else:
            return {}

    @property
    def href(self):
        if self._href:
            return self._href

        # first form field field with an error,
        if self.form_field.errors:
            for subfield in self.form_field:
                if subfield.errors:
                    return "input-" + subfield.name

        # or just the first form field field
        return super().href + "-day"

    @property
    def day(self):
        return self.form_field.day

    @property
    def month(self):
        return self.form_field.month

    @property
    def year(self):
        return self.form_field.year

    def process(self, formdata, data=unset_value):
        '''
        Process incoming data.

        Overrides wtforms.Field.process().

        Filters, process_data and process_formdata are not supported.
        '''
        self.process_errors = []

        # use the FormField to process `formdata` and `data`
        self.form_field.process(formdata, data)

        # make a "fake" raw_data property from the FormField values
        # we need the raw_data property for various validators
        raw_data = {field.name: field.raw_data[0] for field in self.form_field
                    if field.raw_data}
        if not any(raw_data.values()):
            # if all fields were empty we want raw_data to be None-ish
            raw_data = {}

        # the WTForms.Field api expects .raw_data to be a list
        self.raw_data = [raw_data]

        try:
            self.data = datetime.date(**self.form_field.data)
        except (TypeError, ValueError):
            self.data = None
            self.process_errors.append(self.gettext('Not a valid date value'))

        # if the year does not have four digits call the data invalid
        if self.data:
            try:
                digits = len(self.form_field.year.raw_data[0])
            except IndexError:
                digits = 0

            if not digits == 4:
                self.data = None
                self.process_errors.append(self.gettext('Not a valid date value'))

    def post_validate(self, form, validation_stopped):
        # Consumers of a data field might want to highlight individual inputs
        # of the date input component, checking the values of {day, month,
        # year}.errors.  In the case where the validator chain hasn't added
        # errors to any of day month or year, we will add the errors to all the
        # fields so that all inputs get highlighted.
        if self.errors and not any((self.year.errors, self.month.errors, self.day.errors)):
            self.year.errors = self.errors
            self.month.errors = self.errors
            self.day.errors = self.errors



from __future__ import absolute_import

import boto3
import datetime
from dateutil.parser import parse as parse_time
import logging
import mimetypes
import os

import flask

# a bit of a lie here - retains compatibility with consumers that were importing boto2's S3ResponseError from here. this
# is the exception boto3 raises in (mostly) the same situations.
from botocore.exceptions import ClientError as S3ResponseError

from .formats import DATETIME_FORMAT
from .timing import logged_duration_for_external_request as log_external_request

logger = logging.getLogger(__name__)

FILE_SIZE_LIMIT = 5400000  # approximately 5Mb

default_region = "eu-west-1"


class S3(object):
    def __init__(self, bucket_name, region_name=default_region, **kwargs):
        if flask.current_app:
            app = flask.current_app
            if app.env == "development" and app.config.get("DM_S3_ENDPOINT_URL"):
                kwargs.setdefault("endpoint_url", app.config["DM_S3_ENDPOINT_URL"])
        self._resource = boto3.resource("s3", region_name=region_name, **kwargs)
        self._bucket = self._resource.Bucket(bucket_name)

    @property
    def bucket_name(self):
        return self._bucket.name

    def save(self, path, file_, acl='public-read', timestamp=None, download_filename=None,
             disposition_type='attachment'):
        """Save a file in an S3 bucket

        canned ACL list: https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl

        :param path:              location in S3 bucket at which to save the file
        :param file_:             file object to be saved in S3
        :param acl:               S3 canned ACL
        :param timestamp:         Timestamp to set for this file rather than using utcnow
        :param download_filename: Suggested name for a browser to download, part of Content-Disposition header
        :param disposition_type:  Content-Disposition type - e.g. "attachment" or "inline"

        :return: S3 Key
        """
        path = self._normalize_path(path)
        timestamp = timestamp or datetime.datetime.utcnow()
        filesize = get_file_size(file_)

        obj = self._bucket.Object(path)
        extra_kwargs = {}
        if download_filename:
            extra_kwargs["ContentDisposition"] = u'{}; filename="{}"'.format(
                disposition_type,
                # boto/aws can't cope with unicode here, but wants the ultimate result as a `str` in py3, so doing this
                # to strip non-ascii chars..
                str(download_filename).encode("ascii", errors="ignore").decode(),
            )

        log_description = 'file upload [{filepath} of size {filesize} and acl {fileacl}]'
        with log_external_request('S3', log_description) as log_context:
            log_context.update({
                "filepath": path,
                "filesize": filesize,
                "fileacl": acl,
            })

            obj.put(
                ACL=acl,
                Body=file_,
                ContentType=self._get_mimetype(path),
                # using a custom "timestamp" field allows us to manually override it if necessary
                Metadata={"timestamp": timestamp.strftime(DATETIME_FORMAT)},
                **extra_kwargs
            )

        return self._format_key(obj)

    def copy(self, src_bucket, src_key, target_key, acl=None):
        """
        Copy an object that already exists in S3 to a new key in this S3 instance's bucket. The copy will be private by
        default so you need to update the ACL if you wish to have custom access. Doesn't update associated metadata.

        :param src_bucket: The source bucket to look in for the original file.
        :param src_key:    The source key to look for in the source bucket.
        :param target_key: The key to store the copy under in the current bucket.
        :param acl:        If provided, will set the ACL on the key.
        :return:           S3 Key
        :raises:           botocore.exceptions.ClientError if source file doesn't exist.
        """
        if self.path_exists(target_key):
            raise ValueError('Target key already exists in S3.')

        extra_args = {'ACL': acl} if acl else {}

        log_description = 'copy document [{src_bucket}/{src_key} to {target_bucket}/{target_key}{set_acl}]'
        with log_external_request('S3', log_description) as log_context:
            log_context.update({
                "src_bucket": src_bucket,
                "src_key": src_key,
                "target_bucket": self.bucket_name,
                "target_key": target_key,
                "set_acl": f" with '{acl} ACL" if acl else ""
            })

            self._bucket.copy(CopySource={"Bucket": src_bucket, "Key": src_key}, Key=target_key, ExtraArgs=extra_args)

        return self._format_key(self._bucket.Object(target_key))

    @staticmethod
    def _normalize_path(path):
        return path.lstrip('/')

    def path_exists(self, path):
        path = self._normalize_path(path)
        return self._get_key(path) is not None

    def get_signed_url(self, path, expires_in=30):
        """Create a signed S3 document URL

        :param path: S3 object path within the bucket
        :param expires_in: how long the generated URL is valid
                           for, in seconds

        :return: signed URL or ``None`` if object was not found

        """
        path = self._normalize_path(path)
        if self.path_exists(path):
            return self._resource.meta.client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": self._bucket.name,
                    "Key": path,
                },
                ExpiresIn=expires_in,
            )

    def _get_key(self, path):
        path = self._normalize_path(path)
        try:
            obj = self._bucket.Object(path)
            obj.load()
        except S3ResponseError:
            return None
        return obj

    def get_key(self, path):
        path = self._normalize_path(path)
        obj = self._get_key(path)
        return obj and self._format_key(obj)

    def delete_key(self, path):
        path = self._normalize_path(path)
        self._bucket.Object(path).delete()

    def list(self, prefix='', delimiter='', load_timestamps=False):
        """
        return a list of file keys (ordered by last_modified date if load_timestamps is True) from an s3 bucket

        Prefix & Delimiter: http://docs.aws.amazon.com/AmazonS3/latest/dev/ListingKeysHierarchy.html
        :param prefix:          filter by files whose names begin with the prefix
        :param delimiter:       filter out files whose names contain the delimiter
        :param load_timestamps: by default custom timestamps are not loaded as they require an extra API call.
                                if you need to show the timestamp set this to True.
        :return: list
        """
        prefix = self._normalize_path(prefix)

        filtered_objects = self._bucket.objects.filter(Prefix=prefix, Delimiter=delimiter)

        with log_external_request('S3', f'list objects [prefix={prefix}, delimiter={delimiter}]'):
            # Consume the `filtered_objects` generator to memory and prepare for sorting
            objects_for_sorting = [
                self._format_key(obj_s, with_timestamp=load_timestamps)
                for obj_s in filtered_objects
                if not (obj_s.size == 0 and obj_s.key[-1] == '/')
            ]

        return sorted(objects_for_sorting, key=lambda obj_s: (obj_s.get("last_modified") or "", obj_s["path"]))

    def _format_key(self, obj, with_timestamp=True):
        """
        Transform a boto3 s3 Object or ObjectSummary object into a (simpler, implementation-abstracted) dict

        :param obj:            either a boto3 s3 Object or ObjectSummary
        :param with_timestamp: by default our custom timestamps are not loaded as they require an extra API call.
                               if you need to show the timestamp set this to True.
        :return:    dict
        """
        filename, ext = os.path.splitext(os.path.basename(obj.key))

        if with_timestamp and hasattr(obj, "Object"):
            # obj is presumably an ObjectSummary, but we'll need an Object if we want the timestamp, which should get
            # auto-fetched when the attribute is accessed
            obj = obj.Object()

        keydict = {
            'path': obj.key,
            'filename': filename,
            'ext': ext[1:],
            # ObjectSummary has .size, Object has .content_length
            'size': obj.size if hasattr(obj, "size") else obj.content_length,
        }
        if with_timestamp:
            # First look for custom "timestamp" metadata field that is explicitly set by our S3 uploader
            # fall back to AWS's "last_modified" if this doesn't exist
            keydict["last_modified"] = (
                (obj.metadata.get("timestamp") and parse_time(obj.metadata["timestamp"])) or obj.last_modified
            ).strftime(DATETIME_FORMAT)

        return keydict

    @staticmethod
    def _get_mimetype(filename):
        mimetype, _ = mimetypes.guess_type(filename)
        return mimetype


def get_file_size(file_):
    if hasattr(file_, "buffer"):
        # presumably a TextIO object - we want to deal with things on a byte-level though...
        file_ = file_.buffer

    original_pos = file_.tell()
    file_.seek(0, 2)
    size = file_.tell()
    file_.seek(original_pos)
    # see it's like nothing happened, right?

    return size



from contextlib import contextmanager
import inspect
import logging
import sys
import time

from flask import request
from flask.ctx import has_request_context


SLOW_EXTERNAL_CALL_THRESHOLD = 0.25
SLOW_DEFAULT_CALL_THRESHOLD = 0.5


def _logged_duration_default_message(log_context):
    return "Block {} in {{duration_real}}s of real-time".format(
        "executed" if sys.exc_info()[0] is None else "raised {}".format(sys.exc_info()[0].__name__)
    )


def _logged_duration_default_condition(log_context):
    return has_request_context() and (
        getattr(request, "is_sampled", False) or log_context.get("duration_real", 0) > SLOW_DEFAULT_CALL_THRESHOLD
    )


def _logged_duration_default_log_func(logger, message, log_level, log_context):
    final_message = message(log_context) if callable(message) else message
    logger.log(log_level, final_message, exc_info=(sys.exc_info()[0] is not None), extra=log_context)


_logged_duration_default_logger = logging.getLogger(__name__)


@contextmanager
def logged_duration(
    logger=_logged_duration_default_logger,
    message=_logged_duration_default_message,
    log_level=logging.DEBUG,
    condition=_logged_duration_default_condition,
    log_func=_logged_duration_default_log_func,
):
    """
        returns a context manager which will monitor the amount of time spent "inside" its code block and emit a log
        message on exiting block if ``condition`` passes. Uses a ``log_context`` dictionary as the log call's ``extra``
        parameter which will contain the parameters ``duration_real`` and ``duration_process``, each being a float
        duration in seconds. Additional parameters can be added to this ``log_context`` dictionary by annotating them
        to the dictionary which is yielded by the context manager, e.g.::

            with logged_duration(message="Received result {foo} in {duration_real}s") as log_context:
                ...
                log_context["foo"] = do_something()
                ...

        Logging action is triggered whether execution fell out of the block naturally or the code raised an exception.
        If log messages are not desired on exeptions these can always be excluded by a custom ``condition``.

        Can also be used as a function decorator.

        :param logger:    The logger to log the message to. It's best to set this at the very least to give the log
                          reader a clue as to what block of code is being timed.
        :param message:   Message format string to emit. Can be a callable which should return the desired message
                          format string when passed a ``log_context``.
        :param log_level: Numeric log level to use.
        :param condition: Callable accepting a single dictionary argument of the ``log_context``. Should return a
                          boolean signalling whether a log message should be emitted or not. Default condition emits a
                          log if there is a current Flask request with a True `is_sampled` attribute. A
                          condition of True or None will cause logs to *always* be emitted.
        :param log_func:  The actual logging function which will be called if `condition` passes. Arguments passed are:
                          ``logger``, ``message``, ``log_level`` (all verbatim as passed to ``logged_duration``) and
                          ``log_context``.
    """
    original_real_time = time.perf_counter()
    # NOTE this is *process* time, not *thread* time. if multiple threads are running in this process it will include
    # their cpu time too. getting *thread* time will be possible in python 3.7+ (but even then it doesn't work on older
    # macos)
    original_process_time = time.process_time()

    log_context = {}

    try:
        yield log_context
    finally:
        duration_real = time.perf_counter() - original_real_time
        duration_process = time.process_time() - original_process_time

        log_context["duration_real"] = duration_real
        log_context["duration_process"] = duration_process

        if condition in (True, None,) or condition(log_context):
            log_func(logger, message, log_level, log_context)


#
# logged_duration's defaults are exposed here so that a caller is able to defer to the default or simply use a wrapper
# for the default when customizing the call. annotating them *onto* logged_duration itself allows these defaults to be
# used without needing an extra import.
#
logged_duration.default_log_func = _logged_duration_default_log_func  # type: ignore
logged_duration.default_message = _logged_duration_default_message  # type: ignore
logged_duration.default_condition = _logged_duration_default_condition  # type: ignore
# exposing this allows a caller to specify default_logger.getChild(...) as their logger
logged_duration.default_logger = _logged_duration_default_logger  # type: ignore


def exceeds_slow_external_call_threshold(log_context):
    """A public condition that will return True if the duration is above the threshold we have defined as acceptable for
    calls to external services (e.g. Notify, Mailchimp, S3, etc)."""
    return log_context['duration_real'] > SLOW_EXTERNAL_CALL_THRESHOLD


def request_is_sampled(log_context):
    """A public condition that returns True if the request has the X-B3-Sampled flag set in its headers. While this is
    the default condition for logged_duration, exposing it publically allows it to be easily combined with other
    conditions."""
    return has_request_context() and getattr(request, "is_sampled", False)


def exception_in_stack():
    """Return true if we are currently in the process of handling an exception, ie one has been caught in a try block.

    https://docs.python.org/3/library/sys.html#sys.exc_info
    """
    return sys.exc_info()[0] is not None


def different_message_for_success_or_error(success_message, error_message):
    """Can be passed into `logged_duration` as `message=different_message_for_success_or_error(x, y)` in order to
    generate different log messages depending on whether the block completed successfully or raised an exception."""
    return lambda _: success_message if not exception_in_stack() else error_message


def request_context_and_any_of_slow_call_or_sampled_request_or_exception_in_stack(log_context):
    return has_request_context() and (
        exceeds_slow_external_call_threshold(log_context) or request_is_sampled(log_context) or exception_in_stack()
    )


def logged_duration_for_external_request(service, description=None, success_message=None, error_message=None,
                                         logger=None):
    """A default implementation of `logged_duration` to wrap around calls to external services (such as Notify,
    Mailchimp, S3, ...) to generate log messages on these events in a standardised manner.

    This implementation will not log durations outside of a Flask request context (e.g. when running scripts).

    Use to wrap a call to a third-party service like so [note the final () call which accepts additional args]:
    >>> with logged_duration_for_external_request('Notify'):
    >>>     notify_client.send_email('user@email.com')
    """
    if not description:
        # Returns the name of the calling function
        description = inspect.stack()[1].function

    success_message = (
        success_message
        if success_message else
        f'Call to {service} ({description}) executed in {{duration_real}}s'
    )
    error_message = (
        error_message
        if error_message else
        f'Exception from call to {service} ({description}) after {{duration_real}}s'
    )

    return logged_duration(
        message=different_message_for_success_or_error(success_message=success_message, error_message=error_message),
        condition=request_context_and_any_of_slow_call_or_sampled_request_or_exception_in_stack,
        **{'logger': logger} if logger else {}
    )



"""
Support for running on a CloudFoundry platform such as GOV.UK PaaS.
"""

import json
import os
from typing import Dict, List, cast


def get_vcap_services() -> dict:
    return cast(dict, json.loads(os.environ["VCAP_SERVICES"]))


def get_service_by_name_from_vcap_services(vcap_services: Dict[str, List[dict]], name: str) -> dict:
    """Returns the first service from a VCAP_SERVICES json object that has name"""
    for services in vcap_services.values():
        for service in services:
            if service["name"] == name:
                return service

    raise RuntimeError(f"Unable to find service with name {name} in VCAP_SERVICES")



import math
import os
import sys
from flask import jsonify, current_app


from dmutils.timing import logged_duration


def get_version_label(path):
    try:
        path = os.path.join(path, 'version_label')
        with open(path) as f:
            return f.read().strip()
    except IOError:
        return None


def get_disk_space_status(low_disk_percent_threshold=5):
    """Accepts a single parameter that indicates the minimum percentage of disk space which should be free for the
    instance to be considered healthy.

    Returns a tuple containing two items: a status (OK or LOW) indicating whether the disk space remaining on the
    instance is below the threshold and the integer percentage remaining disk space."""
    disk_stats = os.statvfs('/')

    disk_free_percent = int(math.ceil(((disk_stats.f_bfree * 1.0) / disk_stats.f_blocks) * 100))

    return 'OK' if disk_free_percent >= low_disk_percent_threshold else 'LOW', disk_free_percent


class StatusError(Exception):
    """A stub class to use when implementing additional checks for an app's _status endpoint. See API for example.
    When raising a StatusError, make sure that the message passed in uniquely identifies the additional check you are
    performing so that any errors can more easily be tied back to a specific dependency that has failed."""
    message = ''

    def __init__(self, message):
        super().__init__(message)
        self.message = message


def _perform_additional_checks(additional_checks, response_data, error_messages):
    for additional_check in additional_checks:
        try:
            with logged_duration(
                logger=current_app.logger,
                message=(
                    lambda log_context: logged_duration.default_message(log_context) + " ({check_function})"
                ),
                condition=(
                    lambda log_context: logged_duration.default_condition(log_context) or sys.exc_info()[0] is not None
                ),
            ) as log_context:
                log_context["check_function"] = str(additional_check)
                check_result = additional_check()
            response_data.update(check_result)
        except StatusError as e:
            error_messages.append(e.message)


def get_app_status(
    data_api_client=None,
    search_api_client=None,
    ignore_dependencies=False,
    additional_checks=None,
    additional_checks_extended=None,
):
    """Generates output for `_status` endpoints on apps

    :param current_app: The flask `current_app` global.
    :param data_api_client: The app's data_api_client, if used.
    :param search_api_client: The app's search_api_client, if used.
    :param ignore_dependencies: Minimal endpoint checks only (i.e. the app is routable and disk space is fine).
    :param additional_checks: A sequence of callables that return dicts of data to be injected into the final JSON
                              response or raise StatusErrors if they need to log an error that should fail the
                              check (this will cause the endpoint to return a 500). These will be called even if
                              ignore_dependencies=True and should be reserved for checks that the service requires to
                              operate. For example checks to backing services such as persistent datastores or
                              processes.
    :param additional_checks_extended: Similar to `additional_checks`, but only called when ignore_dependencies=False
    :return: A dictionary describing the current state of the app with, at least, a 'status' key with a value of 'ok'
             or 'error'.
    """
    error_messages = []
    response_data = {'status': 'ok'}

    with logged_duration(
        logger=current_app.logger,
        message=(
            lambda log_context: logged_duration.default_message(log_context) + " (get_disk_space_status)"
        ),
    ):
        disk_status = get_disk_space_status()
    response_data['disk'] = f'{disk_status[0]} ({disk_status[1]}% free)'
    if disk_status[0].lower() != 'ok':
        error_messages.append(f'Disk space low: {disk_status[1]}% remaining.')

    if additional_checks:
        _perform_additional_checks(additional_checks, response_data, error_messages)

    if not ignore_dependencies:
        response_data['version'] = current_app.config['VERSION']

        if data_api_client:
            data_api_status = data_api_client.get_status() or {'status': 'n/a'}
            response_data['api_status'] = data_api_status
            if data_api_status['status'].lower() != 'ok':
                error_messages.append('Error connecting to the Data API.')

        if search_api_client:
            search_api_status = search_api_client.get_status() or {'status': 'n/a'}
            response_data['search_api_status'] = search_api_status
            if search_api_status['status'].lower() != 'ok':
                error_messages.append('Error connecting to the Search API.')
        if additional_checks_extended:
            _perform_additional_checks(additional_checks_extended, response_data, error_messages)

    if error_messages:
        current_app.logger.error(
            "Request completed with error_messages = {error_messages}",
            extra={"error_messages": error_messages},
        )
        response_data['status'] = 'error'
        response_data['message'] = error_messages

    return jsonify(**response_data), 200 if not error_messages else 500



from __future__ import unicode_literals

from typing import Iterable, Any

import govuk_country_register
import re
from jinja2 import evalcontextfilter, Markup, escape
from jinja2.nodes import EvalContext


def smartjoin(input: Iterable) -> str:
    list_to_join = list(input)
    if len(list_to_join) > 1:
        return '{} and {}'.format(', '.join(list_to_join[:-1]), list_to_join[-1])
    elif len(list_to_join) == 1:
        return '{}'.format(list_to_join[0])
    else:
        return ''


def format_links(text: str, open_links_in_new_tab: bool = False) -> str:
    """
    Filter that searches a given string (or other string-like object) for any URIs
    and wraps them with either an anchor link or a span, depending on whether the link contains a valid protocol.
    Python3's re library returns matches with type string rather than the arg's type, which
    causes problems for Markup() objects containing tags to be escaped later. Therefore
    we need to cast the matches and formatted links back to the original type before returning the value.
    """
    url_match = re.compile(r"""(
                                (?:https?://|www\.)    # start with http:// or www.
                                (?:[^\s<>"'/?#]+)      # domain doesn't have these characters
                                (?:[^\s<>"']+)         # post-domain part of URL doesn't have these characters
                                [^\s<>,"'\.)]          # no dot or bracket at end
                                )""", re.X)
    matched_urls = [type(text)(substr) for substr in url_match.findall(text)]
    if matched_urls:
        if open_links_in_new_tab:
            link = '<a href="{0}" class="govuk-link" rel="external noreferrer noopener" target="_blank">{0}</a>'
        else:
            link = '<a href="{0}" class="govuk-link" rel="external">{0}</a>'
        text_array = [type(text)(substr) for substr in url_match.split(text)]
        formatted_text_array = []
        for partial_text in text_array:
            if partial_text in matched_urls:
                if partial_text.startswith('www'):
                    url = '{0}'.format(Markup.escape(partial_text))
                else:
                    url = link.format(Markup.escape(partial_text))
                formatted_text_array.append(url)
            else:
                partial_text = Markup.escape(partial_text)
                formatted_text_array.append(partial_text)
        formatted_text = Markup(''.join(formatted_text_array))
        return formatted_text
    else:
        return text


def nbsp(text: str) -> str:
    """Replace spaces with nbsp.

    If you want to use html with this filter you need to pass it in as marksafe
    ie.
    {{ "some text and <html>"|marksafe|nbsp }}"""
    text = escape(text)
    return text.replace(' ', Markup('&nbsp;'))


def capitalize_first(maybe_text: Any) -> Any:
    """If it's a string capitalise the first character, unless it looks like a URL

    :param maybe_text: Could be anything
    :return: If maybe_text is a string it will be returned with an initial capital letter, otherwise unchanged
    """
    if maybe_text and isinstance(maybe_text, str):
        if not maybe_text.startswith('http'):
            return maybe_text[0].capitalize() + maybe_text[1:]
    elif isinstance(maybe_text, (list, tuple)):
        return [capitalize_first(item) for item in maybe_text]

    return maybe_text


# find repeated sequences of '\r\n\', optionally separated by other non-newline whitespace space characters
_multiple_newlines_re = re.compile(r'(\r\n[ \t\f\v]*){2,}')
# find \r\n sequences
_single_newline_re = re.compile(r'(\r\n)')


def replace_newlines_with_breaks(value: str) -> str:
    """Replace newlines in a string with HTML <br> elements"""

    # `escape()` returns Markdown objects in python2
    # We want to cast the output value back into unicode strings
    value = u'{}'.format(escape(value))

    # limit sequences of "\r\n\r\n ..."s to two
    value = _multiple_newlines_re.sub(u'\r\n\r\n', value)

    result = _single_newline_re.sub(u'<br>', value)

    return result


@evalcontextfilter
def preserve_line_breaks(eval_ctx: EvalContext, value: str) -> str:
    result = replace_newlines_with_breaks(value)

    if eval_ctx.autoescape:
        result = Markup(result)

    return result


def sub_country_codes(text: str) -> str:
    """Replace country codes with country common name

    :param text:    text containing country codes in the form 'country:AA'
                    where AA is a two-letter ISO 3166-2 alpha-2 country code
    :return:        text containing country names, or the country code if not found
    """

    def replace_match_with_country_name(match):
        country_code = match["country_code"]
        try:
            return govuk_country_register.to_country(country_code)
        except KeyError:
            return country_code

    try:
        return re.sub(
            r"country:(?P<country_code>[A-Z][A-Z])",
            replace_match_with_country_name,
            text
        )
    except TypeError:
        return text



from typing import Optional

from flask import current_app

__all__ = ["DMCheckboxInput", "DMDateInput", "DMRadioInput", "DMTextInput", "DMTextArea", "DMUnitInput"]


class DMJinjaWidgetBase:

    # we include common template arguments here to avoid repetition
    error = None
    name = None
    hint = None
    question = None
    question_advice = None
    value = None

    def __init__(self, hide_question=False, **kwargs):
        self.__template__ = None

        self.__context__ = {k: getattr(self, k) for k in dir(self) if not k.startswith("_")}
        self.__constructor_kwargs__ = kwargs

        self.__dict__.update(kwargs)

        if hide_question:
            del self.__context__["question"]

    def __call__(self, field, **kwargs):

        # get the template variables
        # precedence (decreasing order)
        # - kwargs to __call__
        # - kwargs to constructor
        # - field values
        # - widget defaults

        context = {}
        context.update(self.__context__)

        for attr in context:
            # we don't want the type from the field because
            # this is just the class name of the field
            if attr == "type":
                continue
            if hasattr(field, attr):
                context[attr] = getattr(field, attr)

        context.update(self.__constructor_kwargs__)
        context.update(kwargs)

        return self._render(context)

    def _render(self, *args, **kwargs):
        # cache the template
        # this cannot be done in __init__ as the flask app may not exist
        if not self.__template__:
            self.__template__ = current_app.jinja_env.get_template(self.__template_file__)

        return self.__template__.render(*args, **kwargs)


class DMSelectionButtonBase(DMJinjaWidgetBase):
    __template_file__ = "toolkit/forms/selection-buttons.html"

    type: Optional[str] = None
    inline = None
    options = None


class DMCheckboxInput(DMSelectionButtonBase):
    type = "checkbox"


class DMDateInput(DMJinjaWidgetBase):
    __template_file__ = "toolkit/forms/date.html"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        del self.__context__["value"]

    def __call__(self, field, **kwargs):
        return super().__call__(field, data=field.value, **kwargs)


class DMRadioInput(DMSelectionButtonBase):
    type = "radio"


class DMTextInput(DMJinjaWidgetBase):
    __template_file__ = "toolkit/forms/textbox.html"


class DMTextArea(DMTextInput):
    large = True

    def __init__(self, max_length_in_words=None, **kwargs):
        if max_length_in_words:
            self.max_length_in_words = max_length_in_words

        super().__init__(**kwargs)


class DMUnitInput(DMTextInput):
    unit = None
    unit_in_full = None
    unit_position = None



from werkzeug.exceptions import Unauthorized


class UnauthorizedWWWAuthenticate(Unauthorized):
    """
    This class exists for back-compatibility. We originally needed it to add support for `www_authenticate` before
    werkzeug added it to Unauthorized. Users should switch to use Unauthorized directly.
    """
    def __init__(self, www_authenticate=None, description=None):
        if not isinstance(www_authenticate, (tuple, list)):
            www_authenticate = (www_authenticate,)
        super().__init__(description=description, www_authenticate=www_authenticate)




# allow importing from dmutils.forms.helpers for backwards compatibility
from .errors import get_errors_from_wtform  # noqa: F401

import typing


def remove_csrf_token(data):
    """Flask-WTF==0.14.2 now includes `csrf_token` in `form.data`, whereas previously wtforms explicitly didn't do
    this. When we pass form data straight through to the API, the API often carries out strict validation and doesn't
    like to see `csrf_token` in the input. So this helper will strip it out of a dict, if it's present.

    Example:
    >>> remove_csrf_token(form.data)
    """
    cleaned_data = {**data}

    if 'csrf_token' in data:
        del cleaned_data['csrf_token']

    return cleaned_data


def govuk_option(
    option: typing.Dict, data: typing.Optional[typing.Union[typing.List[str], str]] = None
) -> typing.Dict:
    """Converts one digitalmarketplace-frontend-toolkit style option (of the radio and checkboxes elements)
    into a format suitable for use with either digitalmarketplace-frontend-toolkit
    templates/macros or govuk-frontend macros.

    Example usage:

    # app.py

        @flask.route("/", methods=["GET", "POST"])
        def view():
            lot_question = {
            option["value"]: option
            for option in govuk_option(
                ContentQuestion(content_loader.get_question(framework_slug, 'services', 'lot')).get('options')
            )
        }
    """
    _data: typing.List[str]
    if data is None:
        _data = []
    elif isinstance(data, str):
        _data = [data]
    elif isinstance(data, list):
        _data = data
    else:
        raise TypeError("`data` must be a string or a list of strings")

    if option:
        # DMp does not require a value for an option, fallback to label if not present
        value = option.get("value", option["label"])
        item = {
            "value": value,
            "text": option['label'],
        }
        if value in _data:
            item["checked"] = True
        if "hint" in option:
            item.update({"hint": {"text": option["hint"]}})
        elif "description" in option:
            item.update({"hint": {"text": option["description"]}})
        return item
    else:
        return {}


def govuk_options(
    options: typing.List[typing.Dict], data: typing.Optional[typing.Union[typing.List[str], str]] = None
) -> typing.List[typing.Dict]:
    """Converts all digitalmarketplace-frontend-toolkit style options (of the radio and checkboxes elements)
    into a format suitable for use with either digitalmarketplace-frontend-toolkit
    templates/macros or govuk-frontend macros.

    Example usage:

    # app.py

    @direct_award_public.route('/<string:framework_family>/choose-lot', methods=("GET", "POST"))
    def choose_lot(framework_family):
        all_frameworks = data_api_client.find_frameworks().get('frameworks')
        framework = framework_helpers.get_latest_live_framework_or_404(all_frameworks, framework_family)

        content_loader.load_messages(framework['slug'], ['advice', 'descriptions'])
        gcloud_lot_messages = content_loader.get_message(framework['slug'], 'advice', 'lots')
        gcloud_lot_messages = {x['slug']: x for x in gcloud_lot_messages}

        lots = list()
        for lot in framework['lots']:
            lot_item = {
                "value": lot['slug'],
                "label": lot['name'],
                "description": gcloud_lot_messages[lot['slug']]['body'],
                "hint": gcloud_lot_messages[lot['slug']].get('advice'),
            }

            lots.append(lot_item)
        return render_template('choose-lot.html',
                            framework_family=framework_family,
                            title="Choose a category",
                            lots=govuk_options(lots))
    """
    return [govuk_option(option, data) for option in options]



from itertools import chain
from random import SystemRandom

from flask import request, current_app


class RequestIdRequestMixin(object):
    """
        A mixin intended for use against a flask Request class, implementing extraction (and partly generation) of
        headers approximately according to the "zipkin" scheme https://github.com/openzipkin/b3-propagation
    """
    # a single class-wide random instance should be good enough for now
    _spanid_random = _traceid_random = SystemRandom()

    @property
    def request_id(self):
        return self.trace_id

    @property
    def trace_id(self):
        """
            The "trace id" (in zipkin terms) assigned to this request. If one was set in the request header, that will
            be used. Failing that, this will be an id we've generated and assigned ourselves.
        """
        if not hasattr(self, "_trace_id"):
            self._trace_id = self._get_first_header(
                current_app.config['DM_TRACE_ID_HEADERS']
            ) or self._get_new_trace_id()
        return self._trace_id

    @property
    def span_id(self):
        """
            The "span id" (in zipkin terms) set in this request's header, if present (None otherwise)
        """
        if not hasattr(self, "_span_id"):
            # note how we don't generate an id of our own. not being supplied a span id implies that we are running in
            # an environment with no span-id-aware request router, and thus would have no intermediary to prevent the
            # propagation of our span id all the way through all our onwards requests much like trace id. and the point
            # of span id is to assign identifiers to each individual request.
            self._span_id = self._get_first_header(current_app.config['DM_SPAN_ID_HEADERS'])
        return self._span_id

    @property
    def parent_span_id(self):
        """
            The "parent span id" (in zipkin terms) set in this request's header, if present (None otherwise)
        """
        if not hasattr(self, "_parent_span_id"):
            self._parent_span_id = self._get_first_header(current_app.config['DM_PARENT_SPAN_ID_HEADERS'])
        return self._parent_span_id

    @property
    def is_sampled(self):
        if not hasattr(self, "_is_sampled"):
            header_value = self._get_first_header(current_app.config['DM_IS_SAMPLED_HEADERS'])
            self._is_sampled = self.debug_flag or (None if header_value is None else header_value == "1")
        return self._is_sampled

    @property
    def debug_flag(self):
        if not hasattr(self, "_debug_flag"):
            header_value = self._get_first_header(current_app.config['DM_DEBUG_FLAG_HEADERS'])
            self._debug_flag = None if header_value is None else header_value == "1"
        return self._debug_flag

    def _get_first_header(self, header_names):
        """
        Returns value of request's first present (and Truthy) header from header_names
        """
        for header_name in header_names:
            if header_name in self.headers and self.headers[header_name]:
                return self.headers[header_name]
        else:
            return None

    def _get_new_trace_id(self):
        bitlen = 128
        return hex(self._traceid_random.randrange(1 << bitlen))[2:].rjust(bitlen // 4, "0")

    def _get_new_span_id(self):
        bitlen = 64
        return hex(self._spanid_random.randrange(1 << bitlen))[2:].rjust(bitlen // 4, "0")

    def get_onwards_request_headers(self):
        """
            Headers to add to any further (internal) http api requests we perform if we want that request to be
            considered part of this "trace id"
        """
        new_span_id = self._get_new_span_id()
        return dict(chain(
            (
                (header_name, self.trace_id)
                for header_name in current_app.config['DM_TRACE_ID_HEADERS']
            ) if self.trace_id else (),
            (
                (header_name, new_span_id)
                for header_name in current_app.config['DM_SPAN_ID_HEADERS']
            ) if self.trace_id else (),
            (
                (header_name, self.span_id)
                for header_name in current_app.config['DM_PARENT_SPAN_ID_HEADERS']
            ) if self.span_id else (),
            (
                (header_name, "1" if self.is_sampled else "0")
                for header_name in current_app.config['DM_IS_SAMPLED_HEADERS']
                # according to zipkin spec we shouldn't propagate the sampling decision if debug_flag is set
            ) if self.is_sampled is not None and not self.debug_flag else (),
            (
                (header_name, "1" if self.debug_flag else "0")
                for header_name in current_app.config['DM_DEBUG_FLAG_HEADERS']
            ) if self.debug_flag is not None else (),
        ))

    def get_extra_log_context(self):
        """
            extra attributes to be made available on a log record based on this request
        """
        return {
            "trace_id": self.trace_id,
            "span_id": self.span_id,
            "parent_span_id": self.parent_span_id,
            # output these as 1|0 strings to match what's easily outputtable by nginx
            "is_sampled": "1" if self.is_sampled else "0",
            "debug_flag": "1" if self.debug_flag else "0",
        }


class ResponseHeaderMiddleware(object):
    def __init__(self, app, trace_id_headers, span_id_headers):
        self.app = app
        self.trace_id_headers = trace_id_headers
        self.span_id_headers = span_id_headers

    def __call__(self, environ, start_response):
        def rewrite_response_headers(status, headers, exc_info=None):
            lower_existing_header_names = frozenset(name.lower() for name, value in headers)
            headers.extend(chain(
                (
                    (header_name, str(request.trace_id),)
                    for header_name in self.trace_id_headers
                    if header_name.lower() not in lower_existing_header_names
                ),
                (
                    (header_name, str(request.span_id),)
                    for header_name in self.span_id_headers
                    if header_name.lower() not in lower_existing_header_names
                ),
            ))

            return start_response(status, headers, exc_info)

        return self.app(environ, rewrite_response_headers)


def init_app(app):
    app.config.setdefault("DM_TRACE_ID_HEADERS", (
        (app.config.get("DM_REQUEST_ID_HEADER") or "DM-Request-ID"),
        (app.config.get("DM_DOWNSTREAM_REQUEST_ID_HEADER") or "X-B3-TraceId"),
    ))
    app.config.setdefault("DM_SPAN_ID_HEADERS", ("X-B3-SpanId",))
    app.config.setdefault("DM_PARENT_SPAN_ID_HEADERS", ("X-B3-ParentSpanId",))
    app.config.setdefault("DM_IS_SAMPLED_HEADERS", ("X-B3-Sampled",))
    app.config.setdefault("DM_DEBUG_FLAG_HEADERS", ("X-B3-Flags",))

    # we do something a little odd here now - back-populate the first value of DM_TRACE_ID_HEADERS back to the
    # DM_REQUEST_ID_HEADER setting, because it turns out that some components (notably the apiclient) depend on that
    # setting
    if app.config.get("DM_TRACE_ID_HEADERS"):
        app.config["DM_REQUEST_ID_HEADER"] = app.config["DM_TRACE_ID_HEADERS"][0]

    # dynamically define this class as we don't necessarily know how request_class may have already been modified by
    # another init_app
    class _RequestIdRequest(RequestIdRequestMixin, app.request_class):
        pass
    app.request_class = _RequestIdRequest
    app.wsgi_app = ResponseHeaderMiddleware(
        app.wsgi_app,
        app.config['DM_TRACE_ID_HEADERS'],
        app.config['DM_SPAN_ID_HEADERS'],
    )



from flask_login import current_user, login_required
from flask import current_app, flash, Markup


ROLES = {
    'buyer': {
        'loginRequiredMessage': Markup("You must log in with a buyer account to see this page."),
    },
    'supplier': {
        'loginRequiredMessage': Markup("You must log in with a supplier account to see this page."),
    },
}


@login_required
def require_login(role):
    """
    Shared function to limit access to a view by role. Can be used with Application.before_request
    or Blueprint.before_request, for example:
        some_blueprint.before_request(functools.partial(require_login, role='buyer'))
    :param role: string, e.g. admin, buyer, supplier
    :return: an unauthorized response, or None if access is allowed (once logged in)
    """
    if current_user.is_authenticated() and current_user.role != role:
        flash(ROLES[role]['loginRequiredMessage'], 'error')
        return current_app.login_manager.unauthorized()



# -*- coding: utf-8 -*-
"""Digital Marketplace normalised exception classes."""


class EmailError(Exception):
    pass


class EmailTemplateError(EmailError):
    pass


class EmailInvalidError(EmailError):
    pass



import unicodecsv


class _StringPipe(object):
    """
    A trivial implementation of something a bit like StringIO but acts more like a pipe than a file,
    flushing its output buffer when read. This way it can be used in lazy iteration for incremental output.
    """
    def __init__(self, initial_value=b""):
        self._contents = initial_value

    def write(self, line):
        self._contents += line

    def read(self):
        retval = self._contents
        self._contents = b""
        return retval


def iter_csv(row_iter, **kwargs):
    pipe = _StringPipe()
    writer = unicodecsv.writer(pipe, **kwargs)
    for row in row_iter:
        writer.writerow(row)
        yield pipe.read()



from dmutils.email.tokens import (
    generate_token, decode_invitation_token, decode_password_reset_token
)

from .exceptions import EmailError
from .dm_notify import DMNotifyClient
from .user_account_email import send_user_account_email



import base64
import json
import struct
from warnings import warn


from datetime import datetime, timedelta

from cryptography import fernet
from flask import current_app

from dmutils.formats import DATETIME_FORMAT
from dmutils.email.helpers import hash_string

ONE_DAY_IN_SECONDS = 86400
SEVEN_DAYS_IN_SECONDS = 604800


def generate_token(json_data, secret_key, namespace):
    """
    Encrypt data using a provided secret_key and namespace.

    The process is as follows:
    * The secret_key is combined with the namespace and hashed using sha256. The namespace ensures that a token for
      `invite-user` cannot be re-used by an attacker on the `reset-password` endpoint. Hashing is used to ensure the
      key conforms to the 32 byte length requirement for Fernet.
    * The combined secret key is used to initialise the Fernet encryption algorithm. Fernet is a wrapper around AES
      that provides HMAC, TTL (time to live), and some quality of life features (url-safe base64 encoding)
      The fernet spec can be viewed here: https://github.com/fernet/spec/blob/master/Spec.md
    * The data is dumped from json and encrypted.
    * The output data is returned as a urlsafe_base64 (https://tools.ietf.org/html/rfc4648#section-5) unicode string.

    Fernet acepts and returns bytes, so call `.encode` before and `.decode` after to convert to strings, to ensure we
    use regular python strings for as much of the code flow as possible

    :param json_data: data to encrypt. Must be json-like blob that `json.dumps` will accept
    :param secret_key: The secret key to encrypt with. No length/content restrictions. Must be a string type.
    :param namespace: The namespace to encrypt with. No length/content restrictions. Must be a string type.
    :return: returns a urlsale_base64 encoded encrypted unicode string.
    :rtype: `unicode`
    """
    secret_key = hash_string(secret_key + namespace)
    data = json.dumps(json_data).encode('utf-8')
    f = fernet.Fernet(secret_key.encode('utf-8'))
    return f.encrypt(data).decode('utf-8')


def _parse_fernet_timestamp(ciphertext):
    """
    Returns utc timestamp embedded in Fernet-encrypted ciphertext, converted to Python datetime object.

    Decryption should be attempted before using this function, as that does cryptographically strong tests on the
    validity of the ciphertext.
    """
    decoded = base64.urlsafe_b64decode(ciphertext)

    try:
        epoch_timestamp = struct.unpack('>Q', decoded[1:9])[0]
    except struct.error as e:
        raise fernet.InvalidToken(e.message)

    return datetime.utcfromtimestamp(epoch_timestamp)


def decode_token(encrypted_data, secret_key, namespace, max_age_in_seconds=ONE_DAY_IN_SECONDS):
    """
    Decrypt data using a provided secret_key, namespace, and TTL (max_age_in_seconds).

    The process is as follows:
    * secret key and fernet are initialised as in `encrypt_data`
    * the data is decrypted and json_dumped
    * the timestamp is pulled out and compared to the max_age_in_seconds to ensure that the token has not expired
    * the original json-able blob is returned, along with the timestamp that it was encrypted at. This timestamp can
      then be used for verifying the authenticity of the request, for example, comparing a password reset token
      against the last time the password was reset to ensure it is not used twice.

    Fernet acepts and returns bytes, so call `.encode` before and `.decode` after to convert to strings, to ensure we
    use regular python strings for as much of the code flow as possible

    :param encrypted_data: data to decrypt. Must be a string type.
    :param secret_key: The secret key you encrypted the data with. Must be a string type.
    :param namespace: The namespace you encrypted the data with. Must be a string type.
    :param max_age_in_seconds: The maximum age of the encrypted data.
    :return: the original encrypted data and the datetime it was encrypted at.
    :rtype: `tuple(json-able, datetime)`
    :raises fernet.InvalidToken: If the secret key and namespace are not able to decrypt the message,
        or max_age_in_seconds has been exceeded.
    """
    encrypted_bytes = encrypted_data.encode('utf-8')
    secret_key = hash_string(secret_key + namespace)
    f = fernet.Fernet(secret_key.encode('utf-8'))

    # this raises fernet.InvalidToken if the key does not match or if TTL is exceeded
    data = f.decrypt(encrypted_bytes, ttl=max_age_in_seconds)

    timestamp = _parse_fernet_timestamp(encrypted_bytes)
    return json.loads(data.decode('utf-8')), timestamp


def decode_password_reset_token(token, data_api_client):
    try:
        decoded, token_timestamp = decode_token(
            token,
            current_app.config["SHARED_EMAIL_KEY"],
            (
                current_app.config.get('RESET_PASSWORD_TOKEN_NS')
                or warn("RESET_PASSWORD_SALT has been renamed RESET_PASSWORD_TOKEN_NS", DeprecationWarning)
                or current_app.config['RESET_PASSWORD_SALT']
            ),
            ONE_DAY_IN_SECONDS,
        )
    except fernet.InvalidToken as e:
        current_app.logger.info("Error changing password: {error}", extra={'error': str(e)})
        return {'error': 'token_invalid'}

    user = data_api_client.get_user(decoded["user"])
    user_last_changed_password_at = datetime.strptime(
        user['users']['passwordChangedAt'],
        DATETIME_FORMAT
    )

    if not user["users"]["active"]:
        current_app.logger.info("Error changing password: target user is not active.")
        return {'error': 'user_inactive'}

    # Check if the token was created before the last password change
    # (allow 1 second leeway for reset tokens generated for the 'Your password was changed' email link)
    if token_timestamp + timedelta(seconds=1) < user_last_changed_password_at:
        current_app.logger.info("Error changing password: Token generated earlier than password was last changed.")
        return {'error': 'token_invalid'}

    if token_timestamp > datetime.utcnow():
        current_app.logger.info("Error changing password: cannot use token generated in the future.")
        return {'error': 'token_invalid'}

    return {
        'user': user['users']['id'],
        'email': user['users']['emailAddress']
    }


def decode_invitation_token(encoded_token):
    try:
        token, _ = decode_token(
            encoded_token,
            current_app.config['SHARED_EMAIL_KEY'],
            (
                current_app.config.get('INVITE_EMAIL_TOKEN_NS')
                or warn("INVITE_EMAIL_SALT has been renamed INVITE_EMAIL_TOKEN_NS", DeprecationWarning)
                or current_app.config['INVITE_EMAIL_SALT']
            ),
            SEVEN_DAYS_IN_SECONDS,
        )
        return token

    except fernet.InvalidToken as error:
        try:
            token, _ = decode_token(
                encoded_token,
                current_app.config['SHARED_EMAIL_KEY'],
                (
                    current_app.config.get('INVITE_EMAIL_TOKEN_NS')
                    or warn("INVITE_EMAIL_SALT has been renamed INVITE_EMAIL_TOKEN_NS", DeprecationWarning)
                    or current_app.config['INVITE_EMAIL_SALT']
                ),
                None,
            )

            current_app.logger.info("Invitation reset attempt with expired token. error {error}",
                                    extra={'error': str(error)})

            return {
                'error': 'token_expired',
                'role': token.get('role') or ('supplier' if token.get('supplier_id') else 'buyer')
            }

        except fernet.InvalidToken as invalid_token_error:
            current_app.logger.info("Invitation reset attempt with invalid token. error {error}",
                                    extra={'error': str(invalid_token_error)})

            return {'error': 'token_invalid'}



from pathlib import PurePath
import re

from werkzeug.routing import BaseConverter, ValidationError


def rewrite_supplier_asset_path(url, assets_url):
    """
    Supplier frontend has an authenticated endpoint for declaration documents: frameworks.download_declaration_document

    Sometimes other apps need to be able to show declaration documents to users who are not the supplier who uploaded
    the document. This function allows them to convert the supplier frontend URL into an assets URL.

    You must ensure that you are only exposing this document to authorised users.
    """
    parts = url.split('/suppliers/assets')
    if len(parts) == 2:
        return assets_url + parts[1]
    return url


class SafePurePathConverter(BaseConverter):
    """
        Like the default :class:`PathConverter`, but it converts
        the output to a :class:`pathlib.PurePath` and ensures it
        contains no path components consisting only of dots.
    """

    regex = "[^/].*?"
    weight = 200

    _disallowed_path_part_re = re.compile(r"\.+$")

    def to_python(self, value):
        pth = PurePath(value)
        if not pth.parts:
            raise ValidationError("All path parts squashed out")

        for part in pth.parts:
            if self._disallowed_path_part_re.match(part):
                raise ValidationError(f"Disallowed path part {part!r}")

        return pth



from datetime import datetime, timedelta
from workdays import workday

from .formats import dateformat, utctoshorttimelongdateformat


def get_publishing_dates(brief):
    APPLICATION_OPEN_DAYS = {'1 week': 7, '2 weeks': 14}
    QUESTIONS_OPEN_DAYS = {'1 week': 2, '2 weeks': 5}
    dates = {}

    if _brief_is_from_api_(brief):
        dates['published_date'] = _get_start_date_from_brief_api(brief)
    elif _brief_is_from_frontend_app_and_published_(brief):
        dates['published_date'] = _get_start_date_from_published_frontend_brief(brief)
    else:
        dates['published_date'] = _get_todays_date()

    length = _get_length_of_brief(brief)

    dates['closing_date'] = dates['published_date'] + timedelta(days=APPLICATION_OPEN_DAYS[length])
    dates['questions_close'] = workday(dates['published_date'], QUESTIONS_OPEN_DAYS[length])
    dates['answers_close'] = workday(dates['closing_date'], -1)
    dates['application_open_weeks'] = length
    dates['closing_time'] = '{d:%I:%M %p}'.format(d=dates['closing_date']).lower()

    return dates


def _brief_is_from_api_(brief):
    return brief.get('publishedAt') and isinstance(brief.get('publishedAt'), datetime)


def _brief_is_from_frontend_app_and_published_(brief):
    return brief.get('publishedAt') and isinstance(brief.get('publishedAt'), str)


def _get_start_date_from_brief_api(brief):
    return brief.get('publishedAt').replace(hour=23, minute=59, second=59, microsecond=0)


def _get_start_date_from_published_frontend_brief(brief):
    return datetime.strptime(brief['publishedAt'][0:10], '%Y-%m-%d') \
        .replace(hour=23, minute=59, second=59, microsecond=0)


def _get_todays_date():
    return datetime.utcnow().replace(hour=23, minute=59, second=59, microsecond=0)


def _get_length_of_brief(brief):
    return brief['requirementsLength'] if brief.get('requirementsLength') else '2 weeks'


def update_framework_with_formatted_dates(framework):
    """Takes a framework record (as returned by eg data_api_client.get_framework) and inserts new keys
    for any date/time values that should be displayed consistently on the frontends.

    This is currently used for important framework lifecycle/application dates."""
    framework.update({
        'clarificationsCloseAt': utctoshorttimelongdateformat(framework['clarificationsCloseAtUTC']),
        'clarificationsPublishAt': utctoshorttimelongdateformat(framework['clarificationsPublishAtUTC']),
        'applicationsCloseAt': utctoshorttimelongdateformat(framework['applicationsCloseAtUTC']),
        'intentionToAwardAt': dateformat(framework['intentionToAwardAtUTC']),
        'frameworkLiveAt': dateformat(framework['frameworkLiveAtUTC']),
        'frameworkExpiresAt': dateformat(framework['frameworkExpiresAtUTC']),
    })




import flask_session
import redis

import dmutils.cloudfoundry as cf


def init_app(app):
    if not app.config.get("SESSION_REDIS"):
        if app.config.get("DM_ENVIRONMENT") == "development":
            app.config["SESSION_REDIS"] = redis.Redis()
        else:
            vcap_services = cf.get_vcap_services()

            redis_service_name = app.config["DM_REDIS_SERVICE_NAME"]
            redis_service = cf.get_service_by_name_from_vcap_services(vcap_services, redis_service_name)

            app.config["SESSION_REDIS"] = redis.from_url(redis_service["credentials"]["uri"])

    app.config["SESSION_USE_SIGNER"] = True
    app.config["SESSION_TYPE"] = 'redis'
    flask_session.Session(app)



import os
import datetime
import re

import fleep

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse  # type: ignore

from .s3 import S3ResponseError, get_file_size


BAD_SUPPLIER_NAME_CHARACTERS = ['#', '%', '&', '{', '}', '\\', '<', '>', '*', '?', '/', '$',
                                '!', "'", '"', ':', '@', '+', '`', '|', '=', ',', '.']

RESULT_LETTER_FILENAME = 'result-letter.pdf'

AGREEMENT_FILENAME = 'framework-agreement.pdf'
SIGNED_AGREEMENT_PREFIX = 'signed-framework-agreement'

SIGNATURE_PAGE_FILENAME = 'signature-page.pdf'
SIGNED_SIGNATURE_PAGE_PREFIX = 'signed-signature-page'

COUNTERPART_FILENAME = "agreement-countersignature.pdf"


def filter_empty_files(files):
    """Remove any empty files from the list.

    :param files: a dictionary of file attachments
    :return: a dictionary of files with all empty files removed

    """
    return {
        key: contents for key, contents in files.items()
        if file_is_not_empty(contents)
    }


def validate_documents(files):
    """Validate document files for size and format

    :param files: a dictionary of file attachments

    :return: a dictionary of errors, where keys match
             the keys from the ``files`` argument and
             values are validator names used to look
             up the message in the question content
             files. If no errors were found an empty
             dictionary is returned.

    """
    errors = {}
    for field, contents in files.items():
        if not file_is_open_document_format(contents):
            errors[field] = 'file_is_open_document_format'
        elif not file_is_less_than_5mb(contents):
            errors[field] = 'file_is_less_than_5mb'

    return errors


def upload_document(uploader, upload_type, documents_url, service, field, file_contents, public=True):
    """Upload the document to S3 bucket and return the document URL

    :param uploader: S3 uploader object
    :param upload_type: Upload type (eg 'documents' or 'submissions')
    :param documents_url: base assets URL used as root for creating the full
                          document URL.
    :param service: service object used to look up service and supplier id
                    for the generated document name
    :param field: name of the service field that the document URL is saved to,
                  used to generate the document name
    :param file_contents: attached file object
    :param public: if True, set file permission to 'public-read'. Otherwise 'bucket-owner-full-control',
                   which is private to the object owner and bucket owner.

    :return: generated document URL or ``False`` if document upload
             failed

    """
    if upload_type not in ('documents', 'submissions',):
        raise ValueError(f"Unexpected upload_type {upload_type!r}")

    file_path = generate_file_name(
        service['frameworkSlug'],
        upload_type,
        service['supplierId'],
        service.get('id'),
        field,
        file_contents.filename
    )

    acl = 'public-read' if public else 'bucket-owner-full-control'

    try:
        uploader.save(file_path, file_contents, acl=acl)
    except S3ResponseError:
        return False

    full_url = urlparse.urljoin(
        documents_url,
        file_path
    )

    return full_url


def upload_declaration_documents(
    uploader, upload_type, documents_url, request_files, section, framework_slug, supplier_id, public=True
):
    # Provide a pseudo 'service' without a Service ID, to construct the filename
    return upload_service_documents(
        uploader, upload_type, documents_url,
        {"frameworkSlug": framework_slug, "supplierId": supplier_id},
        request_files, section, public=public
    )


def upload_service_documents(uploader, upload_type, documents_url, service, request_files, section, public=True):
    if upload_type not in ('documents', 'submissions',):
        raise ValueError(f"Unexpected upload_type {upload_type!r}")

    files = {field: request_files[field] for field in section.get_question_ids(type="upload")
             if field in request_files}
    files = filter_empty_files(files)
    errors = validate_documents(files)

    if errors:
        return None, errors

    if len(files) == 0:
        return {}, {}

    for field, contents in files.items():
        url = upload_document(
            uploader, upload_type, documents_url, service, field, contents,
            public=public)

        if not url:
            errors[field] = 'file_can_be_saved'
        else:
            files[field] = url

    return files, errors


def file_is_not_empty(file_contents):
    return not file_is_empty(file_contents)


def file_is_empty(file_object):
    return len(get_first_128_bytes(file_object)) == 0


def file_is_less_than_5mb(file_):
    return get_file_size(file_) < 5400000


def get_first_128_bytes(file_object):
    """Get the contents of a file like object but replace the pointer when we're finished"""
    position = file_object.tell()
    file_object.seek(0)
    contents = file_object.read(128)
    file_object.seek(position)
    return contents


def get_extension(filename):
    file_name, file_extension = os.path.splitext(filename)
    return file_extension.lower()


def get_possible_extensions_from_format(file_object):
    """
    Pass the first 128 bytes from the file object to fleep for a list of possible extensions

    We don't need to load the whole file into memory to let fleep determine the file type. Apparently the first 128
    bytes are sufficient: https://github.com/floyernick/fleep-py/blob/994bc2c274482d80ab13d89d8f7343eb316d3e44/
        fleep/__init__.py#L55

    :param file_object: File object open as bytes.
    :return: list(str): Possible file formats.
    """
    return fleep.get(get_first_128_bytes(file_object)).extension


def extension_matches_possible_file_formats(file_object, extension):
    """
    Pass the first 128 bytes from the file object to fleep and the supplied extension.

    :param file_object: File object open as bytes.
    :param extension: File object open as bytes.
    :return: boolean: Does extension match guess at format?
    """
    return fleep.get(get_first_128_bytes(file_object)).extension_matches(extension.strip('.'))


def file_is_open_document_format(file_object):
    extension = get_extension(file_object.filename)

    return (
        extension in [".pdf", ".odt", ".ods", ".odp"]
        and extension_matches_possible_file_formats(file_object, extension)
    )


def file_is_pdf(file_object):
    """Checks file extension and format as being PDF."""
    extension = get_extension(file_object.filename)

    return extension == ".pdf" and extension_matches_possible_file_formats(file_object, extension)


def file_is_csv(file_object):
    """Checks file extension as being CSV and checks format does not match something other than CSV."""
    extension = get_extension(file_object.filename)

    return extension == ".csv" and get_possible_extensions_from_format(file_object) == []


def file_is_zip(file_object):
    """Checks file extension and format as being ZIP."""
    extension = get_extension(file_object.filename)

    return extension == ".zip" and extension_matches_possible_file_formats(file_object, extension)


def file_is_image(file_object):
    """Checks file extension and format as being JPG. or PNG. JPEG is normalised to JPG"""
    extension = get_extension(file_object.filename)
    extension = extension if extension != '.jpeg' else '.jpg'

    return extension in [".jpg", ".png"] and extension_matches_possible_file_formats(file_object, extension)


def generate_file_name(framework_slug, upload_type, supplier_id, service_id, field, filename, suffix=None):
    if suffix is None:
        suffix = default_file_suffix()

    ID_TO_FILE_NAME_SUFFIX = {
        'serviceDefinitionDocumentURL': 'service-definition-document',
        'termsAndConditionsDocumentURL': 'terms-and-conditions',
        'sfiaRateDocumentURL': 'sfia-rate-card',
        'pricingDocumentURL': 'pricing-document',
        'modernSlaveryStatement': 'modern-slavery-statement',
        'modernSlaveryStatementOptional': 'modern-slavery-statement',
    }

    if service_id:
        return '{}/{}/{}/{}-{}-{}{}'.format(
            framework_slug,
            upload_type,
            supplier_id,
            service_id,
            ID_TO_FILE_NAME_SUFFIX[field],
            suffix,
            get_extension(filename)
        )

    return '{}/{}/{}/{}-{}{}'.format(
        framework_slug,
        upload_type,
        supplier_id,
        ID_TO_FILE_NAME_SUFFIX[field],
        suffix,
        get_extension(filename)
    )


def default_file_suffix():
    return datetime.datetime.utcnow().strftime("%Y-%m-%d-%H%M")


def get_signed_url(bucket, path, base_url):
    url = bucket.get_signed_url(path)
    if url is not None:
        if base_url is not None:
            url = urlparse.urlparse(url)
            base_url = urlparse.urlparse(base_url)
            url = url._replace(netloc=base_url.netloc, scheme=base_url.scheme).geturl()
        return url


# this method is deprecated
def get_agreement_document_path(framework_slug, supplier_id, document_name):
    return '{0}/agreements/{1}/{1}-{2}'.format(
        framework_slug,
        supplier_id,
        document_name
    )


def get_document_path(framework_slug, supplier_id, bucket_category, document_name):
    return '{0}/{1}/{2}/{2}-{3}'.format(
        framework_slug,
        bucket_category,
        supplier_id,
        document_name
    )


def generate_timestamped_document_upload_path(framework_slug, supplier_id, bucket_category, doc_name):
    """Generates a file path with a timestamp inserted before the file extension"""
    file_name, file_extension = os.path.splitext(doc_name)
    timestamped_file_name = file_name + '-' + datetime.datetime.utcnow().strftime("%Y-%m-%d-%H%M%S")
    timestamped_doc_name = timestamped_file_name + file_extension

    return get_document_path(framework_slug, supplier_id, bucket_category, timestamped_doc_name)


def degenerate_document_path_and_return_doc_name(document_path):
    """Takes a document path and returns just the generic document name

    An inverse function for `get_document_path` which returns what would have been
    the `document_name` param (i.e. removes all framework slug, supplier id and bucket category
    information)
    """
    return re.split(r'/\d+-', document_path)[-1]


def sanitise_supplier_name(supplier_name):
    """Replace ampersands with 'and' and spaces with a single underscore."""
    sanitised_supplier_name = supplier_name.encode("ascii", errors="ignore").decode("ascii").strip()
    sanitised_supplier_name = sanitised_supplier_name.replace(' ', '_').replace('&', 'and')
    for bad_char in BAD_SUPPLIER_NAME_CHARACTERS:
        sanitised_supplier_name = sanitised_supplier_name.replace(bad_char, '')
    while '__' in sanitised_supplier_name:
        sanitised_supplier_name = sanitised_supplier_name.replace('__', '_')
    return sanitised_supplier_name


def generate_download_filename(supplier_id, document_name, supplier_name):
    """
        Used for generating supplier-friendly filenames to set as headers in s3 for files,
        such as framework agreement documents, that suppliers need to download
    """
    return '{}-{}-{}'.format(sanitise_supplier_name(supplier_name), supplier_id, document_name)



from __future__ import absolute_import
import logging
import sys
import re
from os import getpid
import os.path
from threading import get_ident as get_thread_ident
import time

from flask import request, current_app
from flask.ctx import has_request_context

from pythonjsonlogger.jsonlogger import JsonFormatter as BaseJSONFormatter

LOG_FORMAT = '%(asctime)s %(app_name)s %(name)s %(levelname)s ' \
             '%(trace_id)s "%(message)s" [in %(pathname)s:%(lineno)d]'


# fields named in LOG_FORMAT and LOG_FORMAT_EXTRA_JSON_KEYS will always be included in json log output even if
# no such field was supplied in the log record, substituting a None value if necessary.
LOG_FORMAT_EXTRA_JSON_KEYS = (
    "span_id",
    "parent_span_id",
    "is_sampled",
    "debug_flag",
)


logger = logging.getLogger(__name__)


def _common_request_extra_log_context():
    return {
        "method": request.method,
        "url": request.url,
        "endpoint": request.endpoint,
        # pid and thread ident are both available on LogRecord by default, as `process` and `thread`
        # respectively but I don't see a straightforward way of selectively including them only in certain
        # log messages - they are designed to be included when the formatter is being configured. This is why
        # I'm manually grabbing them and putting them in as `extra` here, avoiding the existing parameter names
        # to prevent LogRecord from complaining
        "process_": getpid(),
        # stringifying this as it could potentially be a long that json is unable to represent accurately
        "thread_": str(get_thread_ident()),
    }


def init_app(app):
    app.config.setdefault('DM_LOG_LEVEL', 'INFO')
    app.config.setdefault('DM_APP_NAME', 'none')

    @app.before_request
    def before_request():
        # annotating these onto request instead of flask.g as they probably shouldn't be inheritable from a request-less
        # application context
        request.before_request_real_time = time.perf_counter()
        request.before_request_process_time = time.process_time()

        if getattr(request, "is_sampled", False):
            # emit an early log message to record that the request was received by the app
            current_app.logger.log(
                logging.DEBUG,
                "Received request {method} {url}",
                extra=_common_request_extra_log_context(),
            )

    @app.after_request
    def after_request(response):
        current_app.logger.log(
            logging.ERROR if response.status_code // 100 == 5 else logging.INFO,
            '{method} {url} {status}',
            extra={
                "status": response.status_code,
                "duration_real": (
                    (time.perf_counter() - request.before_request_real_time)
                    if hasattr(request, "before_request_real_time") else None
                ),
                "duration_process": (
                    (time.process_time() - request.before_request_process_time)
                    if hasattr(request, "before_request_process_time") else None
                ),
                **_common_request_extra_log_context(),
            },
        )
        return response

    logging.getLogger().addHandler(logging.NullHandler())
    handlers = [logging.StreamHandler(sys.stdout)]

    del app.logger.handlers[:]

    if app.config.get('DM_PLAIN_TEXT_LOGS'):
        formatter = CustomLogFormatter(LOG_FORMAT)
    else:
        formatter = JSONFormatter(get_json_log_format())

    if app.config.get('DM_LOG_PATH'):
        handlers.append(logging.FileHandler(app.config['DM_LOG_PATH']))

    for handler in handlers:
        configure_handler(handler, app, formatter)

    loglevel = logging.getLevelName(app.config['DM_LOG_LEVEL'])
    loggers = [
        app.logger,
        logging.getLogger('dmutils'),
        logging.getLogger('dmapiclient'),
        logging.getLogger('flask_wtf.csrf'),
    ]
    #  The Elasticsearch client used by the search-api uses urllib3 directly and manages retries itself.
    #  This results in more logging than we want from urllib3, so we don't configure the logger for the search-api.
    if app.config['DM_APP_NAME'] != 'search-api':
        loggers.append(logging.getLogger('urllib3.util.retry'))

    for logger_ in loggers:
        for handler in handlers:
            logger_.addHandler(handler)
        logger_.setLevel(loglevel)
    app.logger.info('Logging configured')


def configure_handler(handler, app, formatter):
    handler.setLevel(logging.getLevelName(app.config['DM_LOG_LEVEL']))
    handler.setFormatter(formatter)
    handler.addFilter(AppNameFilter(app.config['DM_APP_NAME']))
    handler.addFilter(RequestExtraContextFilter())
    handler.addFilter(AppStackLocationFilter("app_", app.root_path))
    if os.environ.get('CF_INSTANCE_INDEX'):
        handler.addFilter(AppInstanceFilter())

    return handler


def get_json_log_format():
    return LOG_FORMAT + "".join(f" %({key})s" for key in LOG_FORMAT_EXTRA_JSON_KEYS)


class AppNameFilter(logging.Filter):
    def __init__(self, app_name):
        self.app_name = app_name

    def filter(self, record):
        record.app_name = self.app_name

        return record


class AppInstanceFilter(logging.Filter):
    def __init__(self):
        self.instance_index = os.environ['CF_INSTANCE_INDEX']

    def filter(self, record):
        record.instance_index = self.instance_index

        return record


class BaseExtraStackLocationFilter(logging.Filter):
    def __init__(self, param_prefix):
        self._param_prefix = param_prefix

    def is_interesting_frame(self, frame):
        raise NotImplementedError

    def enabled_for_record(self, record):
        return True

    def _findCaller(self):
        """
        Cut down copy of Python 3.6.6's logging.Logger.findCaller()

        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = logging.currentframe()
        if f is not None:
            f = f.f_back
        while hasattr(f, "f_code"):
            co = f.f_code
            try:
                # do this in a try block to protect ourselves from faulty is_interesting_frame implementations
                is_interesting_frame = self.is_interesting_frame(f)
            except:  # noqa
                is_interesting_frame = False
            if is_interesting_frame:
                return co.co_filename, f.f_lineno, co.co_name
            f = f.f_back
        return None, None, None

    def filter(self, record):
        if not self.enabled_for_record(record):
            return record

        rv = self._findCaller()

        if rv != (None, None, None,):
            for attr_name, attr_value in zip(("pathname", "lineno", "funcName",), rv):
                setattr(record, self._param_prefix + attr_name, attr_value)

        return record


class AppStackLocationFilter(BaseExtraStackLocationFilter):
    """
        Filter which includes extra information on the first stack frame
        that is from a module *within* the "app", based on the code location's
        filename and a provided base file_path_prefix for the "app"
    """
    def __init__(self, param_prefix, file_path_prefix):
        self._file_path_prefix = os.path.normcase(file_path_prefix)
        super().__init__(param_prefix)

    def is_interesting_frame(self, frame):
        filename = os.path.normcase(frame.f_code.co_filename)
        return os.path.commonpath(
            # regarding abspath here: there is a possibility due to https://bugs.python.org/issue20443 that absolutizing
            # these file paths won't work correctly if our python process has done a chdir. I don't think we generally
            # do that but a better solution might be to outlaw relative imports.
            (os.path.abspath(self._file_path_prefix), os.path.abspath(filename))
        ) == self._file_path_prefix

    def enabled_for_record(self, record):
        return record.levelno >= logging.WARNING or (has_request_context() and getattr(request, "is_sampled", False))


class RequestExtraContextFilter(logging.Filter):
    """
        Filter which will pull extra context from the current request's `get_extra_log_context` method (if present)
        and make this available on log records
    """
    def filter(self, record):
        if has_request_context() and callable(getattr(request, "get_extra_log_context", None)):
            for key, value in request.get_extra_log_context().items():
                setattr(record, key, value)

        return record


class CustomLogFormatter(logging.Formatter):
    """Accepts a format string for the message and formats it with the extra fields"""

    FORMAT_STRING_FIELDS_PATTERN = re.compile(r'\((.+?)\)')

    def add_fields(self, record):
        """Ensure all values found in our `fmt` have non-None entries in `record`"""
        for field in self.FORMAT_STRING_FIELDS_PATTERN.findall(self._fmt):
            # slightly clunky - this is so we catch explicitly-set Nones too and turn them into "-"
            fetched_value = record.__dict__.get(field)
            record.__dict__[field] = fetched_value if fetched_value is not None else "-"

    def format(self, record):
        self.add_fields(record)
        msg = super(CustomLogFormatter, self).format(record)

        try:
            msg = msg.format(**record.__dict__)
        except:  # noqa
            # We know that KeyError, ValueError and IndexError are all possible things that can go
            # wrong here - there is no guarantee that the message passed into the logger is
            # actually suitable to be used as a format string. This is particularly so where an
            # we are logging arbitrary exception that may reference code.
            #
            # We catch all exceptions rather than just those three, because _any_ failure to format the
            # message must not result in an error, otherwise the original log message will never be
            # returned and written to the logs, and that might be important info such as an
            # exception.
            #
            # NB do not attempt to log either the exception or `msg` here, or you will
            # find that too fails and you end up with an infinite recursion / stack overflow.
            logger.info("failed to format log message")
        return msg


class JSONFormatter(BaseJSONFormatter):
    def __init__(self, *args, max_missing_key_attempts=5, **kwargs):
        super().__init__(*args, **kwargs)
        self._max_missing_key_attempts = max_missing_key_attempts

    def process_log_record(self, log_record):
        for key, newkey in (
            ("asctime", "time",),
            ("trace_id", "requestId",),
            ("span_id", "spanId",),
            ("parent_span_id", "parentSpanId",),
            ("app_name", "application",),
            ("is_sampled", "isSampled",),
            ("debug_flag", "debugFlag",),
        ):
            try:
                log_record[newkey] = log_record.pop(key)
            except KeyError:
                pass

        log_record['logType'] = "application"

        missing_keys = {}
        for attempt in range(self._max_missing_key_attempts):
            try:
                log_record['message'] = log_record['message'].format(**log_record, **missing_keys)
            except KeyError as e:
                missing_keys[e.args[0]] = f"{{{e.args[0]}: missing key}}"
            else:
                # execution should only ever reach this point once - when the .format() succeeds
                if missing_keys:
                    logger.warning("Missing keys when formatting log message: {}".format(tuple(missing_keys.keys())))

                break

        else:
            logger.exception("Too many missing keys when attempting to format log message: gave up")

        return log_record



import logging
from typing import Union, cast

import requests
from requests import HTTPError


class DirectPlusError(Exception):
    pass


# See https://directplus.documentation.dnb.com/errorsAndInformationMessages.html
class DUNSNumberNotFound(DirectPlusError):
    status_code = 404


class DUNSNumberInvalid(DirectPlusError):
    status_code = 400


class DirectPlusClient(object):
    """Client to interface with Dun and Bradstreet's Direct Plus API."""

    access_token = None
    protocol = 'https'
    domain = 'plus.dnb.com'
    required_headers = (
        ('Content-Type', 'application/json'),
        ('Accept', 'application/json'),
    )

    def __init__(self, username, password, logger=None):
        self.username = username
        self.password = password
        self.logger = logger if logger else logging.getLogger(__name__)

    def _reset_access_token(self):
        """
        Set the access token and header authorisation parameter for requests to use.
        """
        basic_auth_string = requests.auth._basic_auth_str(self.username, self.password)
        response = self._direct_plus_request(
            'token',
            method='post',
            version='v2',
            payload={'grant_type': 'client_credentials'},
            extra_headers=(('Authorization', basic_auth_string),),
            allow_access_token_reset=False
        )
        self.access_token = response.json().get('access_token')
        self.required_headers += (('Authorization', f'Bearer {self.access_token}'),)

    def _direct_plus_request(
        self,
        endpoint,
        method='get',
        version='v1',
        payload=None,
        extra_headers=(),
        allow_access_token_reset=True
    ):
        """
        Make a request to the Direct Plus API
        """
        if self.access_token is None and allow_access_token_reset is True:
            self._reset_access_token()

        url = f'{self.protocol}://{self.domain}/{version}/{endpoint}'
        headers = {**dict(self.required_headers), **dict(extra_headers)}

        if method != 'get':
            response = getattr(requests, method)(url, headers=headers, json=payload)
        else:
            response = requests.get(url, headers=headers, params=payload)

        if response.status_code == 401 and allow_access_token_reset is True:
            # If access token invalid (401) refresh access token manually
            # and retry request with allow_access_token_reset = False
            self._reset_access_token()
            response = self._direct_plus_request(
                endpoint,
                method=method,
                version=version,
                payload=payload,
                extra_headers=extra_headers,
                allow_access_token_reset=False
            )
        return response

    def get_organization_by_duns_number(self, duns_number: Union[str, int]) -> dict:
        """
        Request a organization by D-U-N-S number from the Direct+ API

        :return: the organization details corresponding to the DUNS number
        :raises DUNSNumberNotFound: if there is no organization for the DUNS number
        :raises DUNSNumberInvalid: if `duns_number` is not a valid DUNS number
        :raises DirectPlusError: if there is an unexpected error response from the API
        """

        # request the company profile, linkage, and executives (v2)
        # https://directplus.documentation.dnb.com/openAPI.html?apiID=cmpelkv2
        response = self._direct_plus_request(
            f'data/duns/{duns_number}', payload={'productId': 'cmpelk', 'versionId': 'v2'}
        )

        try:
            response.raise_for_status()
        except HTTPError as exception:
            try:
                error = response.json()["error"]
            except (ValueError, KeyError) as e:
                error_message = f"Unable to parse Direct+ API response: {response}: {e}"
                raise DirectPlusError(error_message, exception)

            error_message = (
                f"Unable to get supplier for DUNS number '{duns_number}' (HTTP error {response.status_code}): {error}"
            )

            if response.status_code in [DUNSNumberInvalid.status_code, DUNSNumberNotFound.status_code]:
                self.logger.warning(error_message)

                if response.status_code == DUNSNumberInvalid.status_code:
                    raise DUNSNumberInvalid(error_message, exception)
                if response.status_code == DUNSNumberNotFound.status_code:
                    raise DUNSNumberNotFound(error_message, exception)

            else:
                self.logger.error(error_message)
                raise DirectPlusError(error_message, exception)

        try:
            organization = cast(dict, response.json()['organization'])
        except (ValueError, KeyError) as e:
            # this should never happen, so let's propogate it so it is noisy
            self.logger.error(
                f"Unable to parse Direct+ API response: {response}: {e}"
            )
            raise

        return organization



def user_has_role(user, role):
    try:
        return user['users']['role'] == role
    except (KeyError, TypeError):
        return False


class User():
    def __init__(self, user_id, email_address, supplier_id, supplier_name, supplier_organisation_size,
                 locked, active, name, role, user_research_opted_in):
        self.id = user_id
        self.email_address = email_address
        self.name = name
        self.role = role
        self.supplier_id = supplier_id
        self.supplier_name = supplier_name
        self.supplier_organisation_size = supplier_organisation_size
        self.locked = locked
        self.active = active
        self.user_research_opted_in = user_research_opted_in

    def is_authenticated(self):
        return self.is_active()

    def is_active(self):
        return self.active and not self.locked

    def is_locked(self):
        return self.locked

    def is_anonymous(self):
        return False

    def has_role(self, role):
        return self.role == role

    def has_any_role(self, *roles):
        return any(self.has_role(role) for role in roles)

    def get_id(self):
        try:
            return unicode(self.id)  # python 2
        except NameError:
            return str(self.id)  # python 3

    def serialize(self):
        return {
            'id': self.id,
            'name': self.name,
            'emailAddress': self.email_address,
            'supplierId': self.supplier_id,
            'supplierName': self.supplier_name,
            'supplierOrganisationSize': self.supplier_organisation_size,
            'locked': self.locked,
            'userResearchOptedIn': self.user_research_opted_in,
        }

    @staticmethod
    def from_json(user_json):
        user = user_json["users"]
        supplier_id = None
        supplier_name = None
        supplier_organisation_size = None

        if "supplier" in user:
            supplier_id = user["supplier"]["supplierId"]
            supplier_name = user["supplier"]["name"]
            supplier_organisation_size = user["supplier"].get("organisationSize")

        return User(
            user_id=user["id"],
            email_address=user['emailAddress'],
            supplier_id=supplier_id,
            supplier_name=supplier_name,
            supplier_organisation_size=supplier_organisation_size,
            locked=user.get('locked', False),
            active=user.get('active', True),
            name=user['name'],
            role=user['role'],
            user_research_opted_in=user.get('userResearchOptedIn')
        )

    @staticmethod
    def load_user(data_api_client, user_id):
        """Load a user from the API and hydrate the User model"""
        user_json = data_api_client.get_user(user_id=int(user_id))

        if user_json:
            user = User.from_json(user_json)
            if user.is_active():
                return user



from flask import Blueprint, abort

external = Blueprint('external', __name__)


# Buyer frontend
@external.route('/')
def index():
    raise NotImplementedError()


@external.route('/<framework_family>/opportunities/<brief_id>')
def get_brief_by_id(framework_family, brief_id):
    # The Buyer FE currently hardcodes this route to 'digital-outcomes-and-specialists'
    if framework_family != 'digital-outcomes-and-specialists':
        abort(404)
    raise NotImplementedError()


@external.route('/<framework_family>/opportunities')
def list_opportunities(framework_family):
    # The Buyer FE currently hardcodes this route to 'digital-outcomes-and-specialists'
    if framework_family != 'digital-outcomes-and-specialists':
        abort(404)
    raise NotImplementedError()


@external.route('/<framework_family>/services/<service_id>')
def direct_award_service_page(framework_family, service_id):
    # The Buyer FE currently hardcodes this route to 'g-cloud'
    if framework_family != 'g-cloud':
        abort(404)
    raise NotImplementedError()


@external.route('/g-cloud/suppliers')
def suppliers_list_by_prefix():
    raise NotImplementedError()


@external.route('/help')
def help():
    raise NotImplementedError()


@external.route('/terms-and-conditions')
def terms_and_conditions():
    raise NotImplementedError()


@external.route('/privacy-notice')
def privacy_notice():
    raise NotImplementedError()


@external.route('/accessibility-statement')
def accessibility_statement():
    raise NotImplementedError()


@external.route('/cookies')
def cookies():
    raise NotImplementedError()


# Supplier frontend
@external.route('/suppliers')
def supplier_dashboard():
    raise NotImplementedError()


@external.route('/suppliers/supply')
def become_a_supplier():
    raise NotImplementedError()


# Brief Responses frontend
@external.route('/suppliers/opportunities/<brief_id>/responses/start')
def start_brief_response(brief_id):
    raise NotImplementedError()


@external.route('/suppliers/opportunities/<int:brief_id>/responses/result')
def view_response_result(brief_id):
    raise NotImplementedError()


@external.route('/suppliers/opportunities/frameworks/<framework_slug>')
def opportunities_dashboard(framework_slug):
    raise NotImplementedError()


# Admin frontend
@external.route('/admin')
def admin_dashboard():
    raise NotImplementedError()


# User frontend
@external.route('/user/create/<encoded_token>')
def create_user(encoded_token):
    raise NotImplementedError()


@external.route('/user/login')
def render_login():
    raise NotImplementedError()


@external.route('/user/logout', methods=["POST"])
def user_logout():
    raise NotImplementedError()


@external.route('/user/notifications/user-research')
def user_research_consent():
    raise NotImplementedError()


@external.route('/user/change-password')
def change_password():
    raise NotImplementedError()


@external.route('/user/cookie-settings')
def cookie_settings():
    raise NotImplementedError()


# Briefs frontend
@external.route('/buyers')
def buyer_dashboard():
    raise NotImplementedError()


@external.route('/buyers/create')
def create_buyer_account():
    raise NotImplementedError()


@external.route('/buyers/frameworks/<framework_slug>/requirements/<lot_slug>')
def info_page_for_starting_a_brief(framework_slug, lot_slug):
    raise NotImplementedError()


@external.route('/buyers/frameworks/<framework_slug>/requirements/user-research-studios')
def studios_start_page(framework_slug):
    raise NotImplementedError()


@external.route('/buyers/frameworks/<framework_slug>/requirements/<lot_slug>/<brief_id>/supplier-questions')
def supplier_questions(framework_slug, lot_slug, brief_id):
    raise NotImplementedError()



from flask.signals import got_request_exception, request_finished

from gds_metrics import GDSMetrics


class DMGDSMetrics(GDSMetrics):
    """Custom metrics class to prevent metrics endpoint being bound to base application object.

    The config setup on the base application object enforces `require_login` on views there. We don't want this on
    the metrics endpoint.

    This class removes the `app.add_url_rule` call found in the original `init_app`:
    github.com/alphagov/gds_metrics_python/blob/a724653fd7970c47265f6d483541af590deac99b/gds_metrics/__init__.py#L33

    We should then call `add_url_rule` on our metrics blueprint instead (see github.com/alphagov/digitalmarketplace-brief-responses-frontend/blob/a0e89b3c84d6c49393b2264e6b4ca6508e7286d9/app/metrics/__init__.py#L31). # NOQA
    This binds our initialised metrics object's endpoint to the blueprint rather than the base application object.
    """

    def init_app(self, app):
        app.before_request(self.before_request)
        request_finished.connect(self.teardown_request, sender=app)
        got_request_exception.connect(self.handle_exception, sender=app)



import json

from flask import current_app, jsonify
from werkzeug.exceptions import BadRequest


class ValidationError(ValueError):
    @property
    def message(self):
        return self.args[0]


def json_error_handler(e):
    try:
        should_log = False
        should_log = int((hasattr(e, "status_code") and e.status_code) or (hasattr(e, "code") and e.code)) >= 500
    except Exception:
        should_log = True

    if should_log:
        # this is a last ditch effort at outputting a traceback hopefully explaining why we're returning a 5xx.
        # unfortunately flask can't always be trusted to do this on its own. this ends up being a suitable place to
        # do it because these handlers are called by flask while a view-generated exception is still being serviced.
        current_app.logger.warning("Generating error response", exc_info=True, extra={"e": e})

    try:
        # initially we'll try and assume this is an HTTPException of some sort.  for the most part, the default
        # HTTPExceptions render themselves in the desired way if returned as a response. the only change we want to
        # make is to enclose the error description in json.
        response = e.get_response()
        response.set_data(json.dumps({"error": e.description}))
        response.mimetype = current_app.config["JSONIFY_MIMETYPE"]

        return response
    except Exception:
        # either `e` wasn't an HTTPException or something went wrong in trying to jsonify it
        return jsonify(error="Internal error"), 500


def validation_error_handler(validation_error):
    return json_error_handler(BadRequest(description=validation_error.message))



from warnings import warn

from flask import current_app, session, abort, url_for
from . import DMNotifyClient, generate_token, EmailError
from .helpers import hash_string


def send_user_account_email(role, email_address, template_name_or_id, extra_token_data={}, personalisation={}):
    notify_client = DMNotifyClient()

    token_data = {
        'role': role,
        'email_address': email_address
    }
    token_data.update(extra_token_data)

    token = generate_token(
        token_data,
        current_app.config['SHARED_EMAIL_KEY'],
        (
            current_app.config.get('INVITE_EMAIL_TOKEN_NS')
            or warn("INVITE_EMAIL_SALT has been renamed INVITE_EMAIL_TOKEN_NS", DeprecationWarning)
            or current_app.config['INVITE_EMAIL_SALT']
        ),
    )

    link_url = url_for('external.create_user', encoded_token=token, _external=True)

    personalisation_with_link = personalisation.copy()
    personalisation_with_link.update({'url': link_url})

    try:
        notify_client.send_email(
            email_address,
            template_name_or_id=template_name_or_id,
            personalisation=personalisation_with_link,
            reference='create-user-account-{}'.format(hash_string(email_address))
        )
        session['email_sent_to'] = email_address
    except EmailError:
        abort(503, response="Failed to send user creation email.")



"""
Validators for WTForms used in the Digital Marketplace frontend

EmailValidator -- validate that a string is a valid email address
GreaterThan -- compare the values of two fields
FourDigitYear -- validate that a four digit year value is provided
DateValidator -- Error messages for a date input
"""

import datetime
from typing import Any, Dict, Optional

from wtforms.validators import StopValidation, ValidationError

from dmutils.email.helpers import validate_email_address


class EmailValidator:
    """
    Tests whether a string is a valid email address.

    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, message="Please enter a valid email address."):
        self.message = message

    def __call__(self, form, field):
        if not validate_email_address(field.data):
            raise ValidationError(self.message)

        return


class GreaterThan:
    """
    Compares the values of two fields.

    :param fieldname:
        The name of the other field to compare to.

    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, fieldname, message=None):
        self.fieldname = fieldname
        self.message = message

    def __call__(self, form, field):
        try:
            other = form[self.fieldname]
        except KeyError:
            raise ValidationError(
                field.gettext("Invalid field name '%s'." % self.fieldname)
            )
        if field.data is None or other.data is None:
            return
        elif field.data > other.data:
            return

        d = {
            "other_label": hasattr(other, "label")
            and other.label.text
            or self.fieldname,
            "other_name": self.fieldname,
        }
        message = self.message
        if message is None:
            message = field.gettext("Field must be greater than %(other_name)s.")

        raise ValidationError(message % d)


class FourDigitYear:
    """
    Validates that a `DateField`'s year field has a four digit year value.

    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field):
        try:
            digits = len(field.form_field.year.raw_data[0])
        except IndexError:
            digits = 0

        if not digits == 4:
            message = self.message
            if message is None:
                message = field.gettext("Year must be YYYY.")

            raise ValidationError(message)


class DateValidator:
    """Error messages for a date input

    Implements the guidance in [1].

    Instead of taking a `message` DateValidator has two properties,
    `whatever_it_is` and `Whatever_it_is`, which are stand-ins for whatever it
    is the question is asking the user for. `whatever_it_is` is used in the
    middle of a sentence, and should include a participle: for instance, if you
    are asking for the date when a buyer needs a supplier to start work,
    `whatever_it_is` should be `the start date`. `Whatever_it_is` (with a
    capital 'W') is used at the beginning of the sentence, and should not
    include a participle. If `Whatever_it_is` is not provided it will be
    generated from `whatever_it_is` by dropping the first word and then
    capitalising the first letter; in our previous example, for instance,
    `Whatever_it_is` would be `Start date`.

    Alternatively you can provide the error messages desired directly; there
    are three different error messages that can be raised by this validator:

    - nothing_is_entered_error_message
    - date_is_incomplete_error_message
    - date_entered_cant_be_correct_error_message

    The string for `date_is_incomplete_error_message` also includes a
    placeholder for `whatever_is_missing`, in format string syntax.

    [1]:  https://design-system.service.gov.uk/components/date-input/#error-messages
    """

    def __init__(
        self,
        whatever_it_is,
        Whatever_it_is=None,
        *,
        nothing_is_entered_error_message=None,
        date_is_incomplete_error_message=None,
        date_entered_cant_be_correct_error_message=None,
    ):
        self.whatever_it_is = whatever_it_is
        if whatever_it_is and not Whatever_it_is:
            _, rest = whatever_it_is.split(maxsplit=1)
            self.Whatever_it_is = rest[0].upper() + rest[1:]
        else:
            self.Whatever_it_is = Whatever_it_is

        self.nothing_is_entered_error_message = (
            nothing_is_entered_error_message or "Enter {whatever_it_is}"
        )
        self.date_is_incomplete_error_message = (
            date_is_incomplete_error_message
            or "{Whatever_it_is} must include a {whatever_is_missing}"
        )
        self.date_entered_cant_be_correct_error_message = (
            date_entered_cant_be_correct_error_message
            or "{Whatever_it_is} must be a real date"
        )

    def _error(self, error, fields, **kwargs):
        e = ValueError(self.format_error_message(error, **kwargs))
        e.error = error

        # Design System guidance is that if more than one field has an error
        # we should highlight all fields.
        if len(fields) > 1:
            e.fields = {"year", "month", "day"}
        else:
            e.fields = fields

        return e

    def format_error_message(self, error, **kwargs):
        return getattr(self, error + "_error_message").format(
            whatever_it_is=self.whatever_it_is,
            Whatever_it_is=self.Whatever_it_is,
            **kwargs,
        )

    def validate_input(self, raw_data: Dict[str, Any]):
        """Check that year month and day are in `raw_data` and are all integers

        :raises ValueError:
        """
        if not any(raw_data.values()):
            raise self._error(
                "nothing_is_entered",
                set(raw_data.keys()),
            )
        elif not all(raw_data.values()):
            missing = list(sorted(
                name
                for name, value in raw_data.items()
                if not value
            ))
            whatever_is_missing = " and ".join(missing)
            raise self._error(
                "date_is_incomplete",
                set(missing),
                whatever_is_missing=whatever_is_missing,
            )

        def int_or_none(s: Any) -> Optional[int]:
            if not s:
                return None
            try:
                return int(s)
            except ValueError:
                return None

        data = {name: int_or_none(value) for name, value in raw_data.items()}

        if not all(data.values()):
            raise self._error(
                "nothing_is_entered",
                {k for k, v in data.items() if v is None},
            )

    def validate_data(self, data: Dict[str, int]):
        """Check that year month and day in `data` make a valid date

        :raises ValueError:
        """
        year, month, day = data["year"], data["month"], data["day"]

        invalid = set()
        if not (1 <= day and day <= 31):
            invalid.add("day")
        if not (1 <= month and month <= 12):
            invalid.add("month")
        if not (datetime.MINYEAR <= year and year <= datetime.MAXYEAR):
            invalid.add("year")

        if not invalid:
            try:
                datetime.date(year, month, day)
            except ValueError:
                # assume that since we've eliminated other possibilities
                # above it must be that the day given is outside the number
                # of days in the given month and year
                # see https://docs.python.org/3.6/library/datetime.html#datetime.date
                invalid.add("day")

        if invalid:
            raise self._error(
                "date_entered_cant_be_correct",
                invalid,
            )

    def validate_date(self, year, month, day):
        self.validate_input({"year": year, "month": month, "day": day})
        self.validate_data({"year": int(year), "month": int(month), "day": int(day)})

    def __call__(self, form, field):
        """WTForms date validator for DMDateField"""
        try:
            self.validate_date(
                field.form_field.year.data,
                field.form_field.month.data,
                field.form_field.day.data,
            )
        except ValueError as e:
            error_message = str(e)
            error_fields = getattr(e, "fields", {"year", "month", "day"})

            # remove processing errors from field
            field.errors[:] = []

            # add errors to specific form field fields
            for form_field_field in error_fields:
                getattr(field.form_field, form_field_field).errors = [error_message]

            validation_error = StopValidation(error_message)
            validation_error.fields = error_fields
            raise validation_error



from collections import OrderedDict
import os
from types import MappingProxyType

from dmutils import config, logging, proxy_fix, request_id, formats, filters, cookie_probe
from dmutils.errors import api as api_errors, frontend as fe_errors
from dmutils.urls import SafePurePathConverter
import dmutils.session
from flask_wtf.csrf import CSRFError
from werkzeug.exceptions import default_exceptions


frontend_error_handlers = MappingProxyType(OrderedDict((
    (CSRFError, fe_errors.csrf_handler,),
    (400, fe_errors.render_error_page,),
    (401, fe_errors.redirect_to_login,),
    (403, fe_errors.redirect_to_login,),
    (404, fe_errors.render_error_page,),
    (410, fe_errors.render_error_page,),
    (503, fe_errors.render_error_page,),
    (500, fe_errors.render_error_page,),
)))


api_error_handlers = MappingProxyType(OrderedDict(
    (
        (api_errors.ValidationError, api_errors.validation_error_handler,),
    ) + tuple(
        (code, api_errors.json_error_handler) for code in default_exceptions
    ),
))


def init_app(
        application,
        config_object,
        bootstrap=None,
        data_api_client=None,
        db=None,
        login_manager=None,
        search_api_client=None,
        error_handlers=frontend_error_handlers,
):

    application.config.from_object(config_object)
    if hasattr(config_object, 'init_app'):
        config_object.init_app(application)

    application.config.from_object(__name__)

    # all belong to dmutils
    config.init_app(application)
    logging.init_app(application)
    proxy_fix.init_app(application)
    request_id.init_app(application)
    cookie_probe.init_app(application)

    if bootstrap:
        bootstrap.init_app(application)
    if data_api_client:
        data_api_client.init_app(application)
    if db:
        db.init_app(application)
    if login_manager:
        login_manager.init_app(application)
        dmutils.session.init_app(application)
    if search_api_client:
        search_api_client.init_app(application)

    # allow us to use <safepurepath:...> components in route patterns
    application.url_map.converters["safepurepath"] = SafePurePathConverter

    @application.after_request
    def add_header(response):
        # Block sites from rendering our views inside <iframe>, <embed>, etc by default.
        # Individual views may set their own value (e.g. 'SAMEORIGIN')
        if not response.headers.get('X-Frame-Options'):
            response.headers.setdefault('X-Frame-Options', 'DENY')
        return response

    # Make filters accessible in templates.
    application.add_template_filter(filters.capitalize_first)
    application.add_template_filter(filters.format_links)
    application.add_template_filter(filters.nbsp)
    application.add_template_filter(filters.smartjoin)
    application.add_template_filter(filters.preserve_line_breaks)
    application.add_template_filter(filters.sub_country_codes)
    # Make select formats available in templates.
    application.add_template_filter(formats.dateformat)
    application.add_template_filter(formats.datetimeformat)
    application.add_template_filter(formats.datetodatetimeformat)
    application.add_template_filter(formats.displaytimeformat)
    application.add_template_filter(formats.shortdateformat)
    application.add_template_filter(formats.timeformat)
    application.add_template_filter(formats.utcdatetimeformat)
    application.add_template_filter(formats.utctoshorttimelongdateformat)

    @application.context_processor
    def inject_global_template_variables():
        return dict(
            pluralize=pluralize,
            **(application.config['BASE_TEMPLATE_DATA'] or {}))

    for exc_or_code, handler in error_handlers.items():
        application.register_error_handler(exc_or_code, handler)


def pluralize(count, singular, plural):
    return singular if count == 1 else plural


def get_extra_files(paths):
    for path in paths:
        for dirname, dirs, files in os.walk(path):
            for filename in files:
                filename = os.path.join(dirname, filename)
                if os.path.isfile(filename):
                    yield filename



from odf.element import Element
from odf.namespaces import OFFICENS
from odf.opendocument import OpenDocumentSpreadsheet
from odf.style import Style
from odf.table import Table, TableColumn, TableRow, TableCell, CoveredTableCell
from odf.text import P, A  # noqa (used by frontend apps)


class Row(object):
    def __init__(self, **kwargs):
        self._row = TableRow(**kwargs)

    def write_cell(self, value, **kwargs):
        if "numbercolumnsspanned" in kwargs or "numberrowsspanned" in kwargs:
            kwargs.setdefault("numberrowsspanned", "1")
            kwargs.setdefault("numbercolumnsspanned", "1")

        cell = TableCell(**kwargs)
        cell.setAttrNS(OFFICENS, "value-type", "string")

        if isinstance(value, Element):
            para = P()
            para.addElement(value)
            cell.addElement(para)
        else:
            for line in value.split("\n"):
                cell.addElement(P(text=line))

        self._row.addElement(cell)

    def write_cells(self, cells, **kwargs):
        for cell in cells:
            self.write_cell(value=cell, **kwargs)

    def write_covered_cell(self):
        self._row.addElement(CoveredTableCell())


class Sheet(object):
    def __init__(self, name):
        self._table = Table(name=name)
        self._rows = {}

    def create_row(self, name, **kwargs):
        """Create an empty row to manually insert cells"""
        self._rows[name] = Row(**kwargs)
        self._table.addElement(self._rows[name]._row)

        return self._rows[name]

    def write_row(self, name, cells, row_styles={}, cell_styles={}):
        """Create a new row and populate it with the given cells"""
        row = self.create_row(name, **row_styles)
        row.write_cells(cells, **cell_styles)

    def get_row(self, name):
        return self._rows[name]

    def create_column(self, **kwargs):
        column = TableColumn(**kwargs)

        self._table.addElement(column)

    def read_cell(self, x, y):
        try:
            cell = self._table.getElementsByType(TableRow)[y].childNodes[x]
        except IndexError:
            return ''

        result = []

        for element in cell.childNodes:
            result.append("".join([v.data for v in element.childNodes]))

        return "\n".join(result)


class SpreadSheet(object):
    def __init__(self):
        self._document = OpenDocumentSpreadsheet()
        self._sheets = {}

    def sheet(self, name):
        if name not in self._sheets:
            self._sheets[name] = Sheet(name)
            self._document.spreadsheet.addElement(self._sheets[name]._table)

        return self._sheets[name]

    def add_style(self, name, family, styles, **kwargs):
        style = Style(name=name, family=family, **kwargs)

        for v in styles:
            style.addElement(v)

        self._document.automaticstyles.addElement(style)

    def add_font(self, fontface):
        self._document.fontfacedecls.addElement(fontface)

    def save(self, buf):
        return self._document.save(buf)



# -*- coding: utf-8 -*-
"""Email helpers."""
import base64
import hashlib
import re


# List of characters that could be used to separate email addresses,
# in decreasing precedence order.
# Space (" ") should almost always be at the end of this list, otherwise
# it can cause weird issues when there are separators surrounded by space.
EMAIL_ADDRESS_SEPARATORS = [";", ",", "/", "&", " "]

# Largely copied from https://github.com/alphagov/notifications-utils/blob/\
#   67889886ec1476136d12e7f32787a7dbd0574cc2/notifications_utils/recipients.py
#
# regexes for use in validate_email_address.
# invalid local chars - whitespace, quotes and apostrophes, semicolons and colons, GBP sign
# Note: Normal apostrophe eg `Firstname-o'surname@domain.com` is allowed.
INVALID_LOCAL_CHARS = r"\s\",;:@£“”‘’"
email_regex = re.compile(r'^[^{}]+@([^.@][^@]+)$'.format(INVALID_LOCAL_CHARS))
hostname_part = re.compile(r'^(xn-|[a-z0-9]+)(-[a-z0-9]+)*$', re.IGNORECASE)
tld_part = re.compile(r'^([a-z]{2,63}|xn--([a-z0-9]+-)*[a-z0-9]+)$', re.IGNORECASE)


def hash_string(string):
    """Hash a given string."""
    m = hashlib.sha256(str(string).encode('utf-8'))
    return base64.urlsafe_b64encode(m.digest()).decode('utf-8')


def validate_email_address(email_address):
    """Return True if `email_address` is valid, otherwise returns False"""

    # Largely a straight copy from https://github.com/alphagov/notifications-utils/blob/\
    #   67889886ec1476136d12e7f32787a7dbd0574cc2/notifications_utils/recipients.py#L439 onwards so that we have
    # validity-parity with Notify and minimise nasty surprises once we attempt to send an email to this address via
    # Notify and only find out it won't be accepted once it's too late to give the user a sane validation message

    # almost exactly the same as by https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py,
    # with minor tweaks for SES compatibility - to avoid complications we are a lot stricter with the local part
    # than neccessary - we don't allow any double quotes or semicolons to prevent SES Technical Failures
    email_address = (email_address or "").strip()
    match = re.match(email_regex, email_address)

    # not an email
    if not match:
        return False

    hostname = match.group(1)
    # don't allow consecutive periods in domain names
    if '..' in hostname:
        return False

    # idna = "Internationalized domain name" - this encode/decode cycle converts unicode into its accurate ascii
    # representation as the web uses. '例え.テスト'.encode('idna') == b'xn--r8jz45g.xn--zckzah'
    try:
        hostname = hostname.encode('idna').decode('ascii')
    except UnicodeError:
        return False

    parts = hostname.split('.')

    if len(hostname) > 253 or len(parts) < 2:
        return False

    for part in parts:
        if not part or len(part) > 63 or not hostname_part.match(part):
            return False

    # if the part after the last . is not a valid TLD then bail out
    if not tld_part.match(parts[-1]):
        return False

    return True


def get_email_addresses(string, separators=EMAIL_ADDRESS_SEPARATORS):
    """
    Returns a list of email addresses from a string.

    It does not validate the email addresses.

    @param separators   list of characters that could be
                        used to separate email addresses

    >>> from dmutils.email.helpers import get_email_addresses
    >>> get_email_addresses("bob@blob.com ")
    ['bob@blob.com']
    >>> get_email_addresses("bob@blob.com; bob.blob@job.com")
    ['bob@blob.com', 'bob.blob@job.com']
    >>> get_email_addresses("bob@blob.com /  bob.blob@job.com")
    ['bob@blob.com', 'bob.blob@job.com']
    >>> get_email_addresses("bob@invalid;bob.blob@job.com")
    ['bob@invalid', 'bob.blob@job.com']
    >>> get_email_addresses("bob@blob;bob.blob@job.com;bob@blob..invalid")
    ['bob@blob', 'bob.blob@job.com', 'bob@blob..invalid']
    >>> get_email_addresses("bob@blob & bob.blob@job.com")
    ['bob@blob', 'bob.blob@job.com']
    """

    addresses = [string]
    for sep in separators:
        if sep in string:
            addresses = string.split(sep)
            break  # earlier separators take precedence

    addresses = (s.strip() for s in addresses)
    addresses = [s for s in addresses if s]  # remove empty strings that can be left in by the algorithm

    return addresses



from flask import redirect, render_template, url_for, flash, session, request, current_app
from jinja2.exceptions import TemplateNotFound

from dmutils import cookie_probe


def csrf_handler(csrf_error):
    """
    :param csrf_error: CSRF exception instance
    :param session: Flask session instance
    :param request: Flask request instance
    :param logger: app logger instance
    :return: redirect to login with flashed error message
    """
    # The "cookie probe" check is performed here because the first point at which a typical session will suffer
    # from cookies not working is when a CSRF token fails to validate
    if cookie_probe.expected_probe_cookie_missing():
        return render_error_page(
            status_code=400,
            error_message="This feature requires cookies to be enabled for correct operation",
        )
    elif 'user_id' not in session:
        current_app.logger.info(
            'csrf.session_expired: Redirecting user to log in page',
            extra={'error': csrf_error.description},
        )
    else:
        current_app.logger.info(
            'csrf.invalid_token: Aborting request, user_id: {user_id}',
            extra={'user_id': session['user_id'], 'error': csrf_error.description}
        )

    flash('Your session has expired. Please log in again.', "error")
    return redirect_to_login(csrf_error)


def redirect_to_login(e):
    if request.method == 'GET':
        return redirect(url_for('external.render_login', next=request.path))
    else:
        return redirect(url_for('external.render_login'))


def render_error_page(e=None, status_code=None, error_message=None):
    """
    Either an exception or a status code must be supplied.
    :param e: exception instance, e.g. Forbidden()
    :param status_code: int status code, e.g. 403
    :return: Flask render_template response. The error templates must be present in the app (either
    copied from the FE Toolkit, or a custom template for that app).
    """
    orig_e, orig_status_code, orig_error_message = e, status_code, error_message

    if not (e or status_code):
        # Something's gone wrong! To save going in an endless error loop let's just render a 500
        status_code = 500

    template_map = {
        400: "errors/400.html",
        404: "errors/404.html",
        410: "errors/410.html",
        500: "errors/500.html",
        503: "errors/500.html",
    }

    if not status_code:
        # Handle exceptions with .status_code, not .code (e.g. dmapiclient.HTTPError)
        if hasattr(e, 'status_code'):
            status_code = e.status_code
        elif hasattr(e, 'code'):
            status_code = e.code
        else:
            # Map unknown status codes to 500
            status_code = 500

    if status_code not in template_map:
        status_code = 500

    if status_code > 499:
        # this is a last ditch effort at outputting a traceback hopefully explaining why we're returning a 5xx.
        # unfortunately flask can't always be trusted to do this on its own. this ends up being a suitable place to
        # do it because these handlers are called by flask while a view-generated exception is still being serviced.
        # we're still emitting a log message even if there's no current exception because the fact that there is no
        # exception *itself* tells us something about why we ended up at a 5xx.
        current_app.logger.warning("Rendering error page", exc_info=True, extra={
            "e": orig_e,
            "status_code": orig_status_code,
            "error_message": orig_error_message,
        })

    try:
        # Try app error templates first
        return render_template(template_map[status_code], error_message=error_message), status_code
    except TemplateNotFound:
        # Fall back to toolkit error templates
        return render_template("toolkit/{}".format(template_map[status_code]), error_message=error_message), status_code



from jinja2.sandbox import SandboxedEnvironment

from .filters import format_links, nbsp, smartjoin, sub_country_codes
from .formats import dateformat, datetimeformat, datetodatetimeformat, shortdateformat


CUSTOM_FILTERS = {
    'format_links': format_links,
    'nbsp': nbsp,
    'smartjoin': smartjoin,
    'dateformat': dateformat,
    'datetimeformat': datetimeformat,
    'shortdateformat': shortdateformat,
    'datetodatetimeformat': datetodatetimeformat,
    'sub_country_codes': sub_country_codes,
}


class DMSandboxedEnvironment(SandboxedEnvironment):
    """DigitalMarketplace environment with filters."""

    def __init__(self, *args, **kwargs):
        super(DMSandboxedEnvironment, self).__init__(*args, **kwargs)
        self.filters.update(CUSTOM_FILTERS)



'''
This module includes mixins that are designed to make it easier to use WTForms
with the Digital Marketplace frontend toolkit.

For example:

    >>> from wtforms import Form, StringField
    >>>
    >>> # include the mixin in a new class
    >>> class DMStringField(DMFieldMixin, StringField):
    ...     pass
    >>>
    >>> # create a form with our new field class
    >>> class TextForm(Form):
    ...     text = DMStringField('Text field', hint='Type text here.')
    >>>
    >>> form = TextForm()
    >>>
    >>> # our new field has all the benefits of DMFieldMixin
    >>> form.text.name
    'text'
    >>> form.text.question
    'Text field'
    >>> form.text.hint
    'Type text here.'

For more examples of these mixins in action, see `dmutils/forms/dm_fields.py`.

For more about mixins in general, see this `online article`_,
or the very excellent `Python Cookbook`_ by David Beazley.

.. _online article:
    https://zonca.github.io/2013/04/simple-mixin-usage-in-python.html
.. _Python Cookbook:
    http://dabeaz.com/cookbook.html
'''

from copy import copy

from wtforms.compat import text_type


class DMFieldMixin:
    '''
    A mixin designed to make it easier to use WTForms
    with the frontend toolkit. The idea is that the
    properties mirror the definitions that the toolkit
    template macros expect.

    It adds the following features:
        - a `hint` property that can be set in the initialiser
        - a `question` property that contains the label text
        - a `value` property for the data that is displayed
        - an `error` property for the field validation error

    Derived classes which include this mixin should have a
    subclass of `wtforms.Field` in their base classes.
    '''
    def __init__(self, label=None, validators=None, hint=None, question_advice=None, **kwargs):
        super().__init__(label=label, validators=validators, **kwargs)

        self._href = kwargs.get("id")

        if hint:
            self.hint = hint
        if question_advice:
            self.question_advice = question_advice

        # wtforms.Field overwrites self.type on init
        # if we want to specify it on a subclass
        # this line will bring it back
        self.type = getattr(self.__class__, 'type', self.type)

    @property
    def href(self):
        return self._href or f"input-{self.id}"

    @property
    def question(self):
        return self.label.text

    @question.setter
    def question(self, value):
        self.label.text = value

    @property
    def value(self):
        return self._value()

    @property
    def error(self):
        try:
            return self.errors[0]
        except IndexError:
            return None


class DMSelectFieldMixin:
    '''
    A Digital Marketplace wrapper for selection fields.

    The `options` argument for the constructor should be a dictionary with
    `value`, label`, and `description` keys.

    The `options` attribute is the choices in a format suitable for the
    frontend toolkit.

    For backwards compatibility, the constructor will accept a `choices`
    keyword argument. However, if both `options` and `choices` are specified
    then `options` will take precedence.
    '''
    def __init__(self, label=None, validators=None, coerce=text_type, options=None, **kwargs):
        super().__init__(label, validators=validators, coerce=coerce, **kwargs)
        if options:
            self.options = copy(options)

    @property
    def choices(self):
        return [(option['value'], option['label']) for option in self.options] if self.options else []

    @choices.setter
    def choices(self, value):
        if value is None:
            self.options = None
        else:
            self.options = []
            for value, label in value:
                self.options.append(
                    {
                        'label': label,
                        'value': value,
                    }
                )

    @property
    def href(self):
        # This should be the id for the first option
        return self._href or super().href + "-1"

    @property
    def value(self):
        if not self.data or self.data == "None":
            return None
        else:
            return self.data



# -*- coding: utf-8 -*-
"""Digital Marketplace MailChimp integration."""

from json.decoder import JSONDecodeError
from hashlib import md5
from logging import Logger
from typing import Callable, Iterator, Mapping, Sequence, Union, cast

from requests.exceptions import RequestException, HTTPError

from mailchimp3 import MailChimp
from mailchimp3.mailchimpclient import MailChimpError

from dmutils.timing import logged_duration_for_external_request as log_external_request

PAGINATION_SIZE = 1000


def get_response_from_exception(exc):
    if isinstance(exc, RequestException):
        try:
            return exc.response.json()
        except (AttributeError, ValueError, JSONDecodeError):
            pass
    if isinstance(exc, MailChimpError):
        return exc.args[0]
    return {}


class DMMailChimpClient(object):

    def __init__(
        self,
        mailchimp_username: str,
        mailchimp_api_key: str,
        logger: Logger,
        retries: int = 0,
    ):
        self._client = MailChimp(mc_user=mailchimp_username, mc_api=mailchimp_api_key, timeout=25)
        self.logger = logger
        self.retries = retries

    @staticmethod
    def get_email_hash(email_address: Union[str, bytes]) -> str:
        """md5 hashing of lower cased emails has been chosen by mailchimp to identify email addresses"""
        formatted_email_address = str(email_address.lower()).encode('utf-8')
        return md5(formatted_email_address).hexdigest()

    def timeout_retry(self, method: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            for i in range(1 + self.retries):
                try:
                    with log_external_request(service='Mailchimp'):
                        return method(*args, **kwargs)
                except HTTPError as e:
                    exception = e
                    if exception.response.status_code == 504:
                        continue
                    raise exception
            raise exception

        return wrapper

    def create_campaign(self, campaign_data: Mapping) -> Union[str, bool]:
        try:
            with log_external_request(service='Mailchimp'):
                campaign = self._client.campaigns.create(campaign_data)
            return cast(str, campaign['id'])
        except (RequestException, MailChimpError) as e:
            self.logger.error(
                "Mailchimp failed to create campaign for '{campaign_title}'".format(
                    campaign_title=campaign_data.get("settings", {}).get("title")
                ),
                extra={
                    "error": str(e),
                    "mailchimp_response": get_response_from_exception(e),
                },
            )
        return False

    def set_campaign_content(self, campaign_id: str, content_data: Mapping):
        try:
            with log_external_request(service='Mailchimp'):
                return self._client.campaigns.content.update(campaign_id, content_data)
        except (RequestException, MailChimpError) as e:
            self.logger.error(
                "Mailchimp failed to set content for campaign id '{0}'".format(campaign_id),
                extra={
                    "error": str(e),
                    "mailchimp_response": get_response_from_exception(e),
                },
            )
        return False

    def send_campaign(self, campaign_id: str):
        try:
            with log_external_request(service='Mailchimp'):
                self._client.campaigns.actions.send(campaign_id)
            return True
        except (RequestException, MailChimpError) as e:
            self.logger.error(
                "Mailchimp failed to send campaign id '{0}'".format(campaign_id),
                extra={
                    "error": str(e),
                    "mailchimp_response": get_response_from_exception(e),
                }
            )
        return False

    def subscribe_new_email_to_list(self, list_id: str, email_address: str):
        """
        Will subscribe email address to list if they do not already exist in that list else do nothing.
        Possible return values:
        True:          User error, e.g. already on list, fake/invalid email (status 4xx)
        False:         Unexpected error, e.g. cannot connect to Mailchimp (status 5xx)
        deleted_user:  Mailchimp can't subscribe a deleted user (status 400)
        anything else: User successfully subscribed (status 200)
        """
        hashed_email = self.get_email_hash(email_address)
        try:
            with log_external_request(service='Mailchimp'):
                resp = self._client.lists.members.create_or_update(
                    list_id,
                    hashed_email,
                    {
                        "email_address": email_address,
                        "status_if_new": "subscribed"
                    }
                )
                resp.update({"status": "success", "error_type": None, "status_code": 200})
                return resp
        except (RequestException, MailChimpError) as e:
            # Some errors we don't care about but do want to log. Find and log them here.
            response = get_response_from_exception(e)
            if "looks fake or invalid, please enter a real email address." in response.get("detail", ""):
                # As defined in mailchimp API documentation, this particular error message may arise if a user has
                # requested mailchimp to never add them to mailchimp lists. In this case, we resort to allowing a
                # failed API call (but log) as a user of this method would unlikely be able to do anything as we have
                # no control over this behaviour.
                self.logger.warning(
                    f"Expected error: Mailchimp failed to add user ({hashed_email}) to list ({list_id}). "
                    "API error: The email address looks fake or invalid, please enter a real email address.",
                    extra={"error": str(e), "mailchimp_response": response}
                )
                return {"status": "error", "error_type": "invalid_email", "status_code": 400}
            elif 'is already a list member.' in response.get("detail", ""):
                # If a user is already a list member we receive a 400 error as documented in the tests for this error
                self.logger.warning(
                    f"Expected error: Mailchimp failed to add user ({hashed_email}) to list ({list_id}). "
                    "API error: This email address is already subscribed.",
                    extra={"error": str(e), "mailchimp_response": response}
                )
                return {"status": "error", "error_type": "already_subscribed", "status_code": 400}
            elif 'The contact must re-subscribe to get back on the list.' in response.get('detail', ''):
                # User has been deleted and cannot be programmatically resubscribed
                self.logger.warning(
                    f"Expected error: Mailchimp cannot automatically subscribe user ({hashed_email}) to list "
                    f"({list_id}) as the user has been permanently deleted.",
                    extra={"error": str(e), "mailchimp_response": response}
                )
                return {"status": "error", "error_type": "deleted_user", "status_code": 400}
            elif response.get('status') == 400:
                # Some other validation error
                self.logger.warning(
                    f"Expected error: Mailchimp failed to add user ({hashed_email}) to list ({list_id}). "
                    "API error: The email address was invalid.",
                    extra={"error": str(e), "mailchimp_response": response}
                )
                return {"status": "error", "error_type": "invalid_email", "status_code": 400}

            # Otherwise this was an unexpected error and should be logged as such
            self.logger.error(
                f"Mailchimp failed to add user ({hashed_email}) to list ({list_id})",
                extra={"error": str(e), "mailchimp_response": response}
            )
            return {"status": "error", "error_type": "unexpected_error", "status_code": 500}

    def subscribe_new_emails_to_list(self, list_id: str, email_addresses: str) -> bool:
        success = True
        for email_address in email_addresses:
            with log_external_request(service='Mailchimp'):
                if not self.subscribe_new_email_to_list(list_id, email_address):
                    success = False
        return success

    def get_email_addresses_from_list(
            self,
            list_id: str,
            pagination_size: int = 100,
            **query_parameters,
    ) -> Iterator[str]:
        """
        See https://mailchimp.com/developer/marketing/api/list-members/list-members-info/ for possible query parameters.
        """
        offset = 0
        while True:
            member_data = self.timeout_retry(
                self._client.lists.members.all
            )(list_id, count=pagination_size, offset=offset, **query_parameters)
            if not member_data.get("members", None):
                break
            offset += pagination_size

            yield from [member['email_address'] for member in member_data['members']]

    def get_lists_for_email(self, email_address: str) -> Sequence[Mapping]:
        """
            Returns a sequence of all lists the email_address has an association with (note: even if that association is
            "unsubscribed" or "cleaned").
        """
        with log_external_request(service='Mailchimp'):
            return tuple(
                {
                    "list_id": mailing_list["id"],
                    "name": mailing_list["name"],
                } for mailing_list in self._client.lists.all(get_all=True, email=email_address)["lists"]
            )

    def permanently_remove_email_from_list(self, email_address: str, list_id: str) -> bool:
        """
            Permanently (very permanently) erases all trace of an email address from a given list
        """
        hashed_email = self.get_email_hash(email_address)
        try:
            with log_external_request(service='Mailchimp'):
                self._client.lists.members.delete_permanent(
                    list_id=list_id,
                    subscriber_hash=hashed_email,
                )
            return True
        except (RequestException, MailChimpError) as e:
            self.logger.error(
                f"Mailchimp failed to permanently remove user ({hashed_email}) from list ({list_id})",
                extra={"error": str(e), "mailchimp_response": get_response_from_exception(e)},
            )
        return False



# -*- coding: utf-8 -*-
from datetime import datetime
import re
from typing import Union

import pytz

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
DATE_FORMAT = "%Y-%m-%d"
DISPLAY_MONTH_YEAR_FORMAT = "%B %Y"
DISPLAY_SHORT_DATE_FORMAT = '%-d %B'
DISPLAY_NO_DAY_DATE_FORMAT = '%-d %B %Y'
DISPLAY_DATE_FORMAT = '%A %-d %B %Y'
DISPLAY_TIME_FORMAT = '%H:%M:%S'
DISPLAY_TIME_TZ_FORMAT = '%-I:%M%p %Z'
DISPLAY_DATETIME_FORMAT = '%A %-d %B %Y at %I:%M%p %Z'
DISPLAY_SHORTTIME_LONGDATE_FORMAT = '%-I%p %Z, %A %-d %B %Y'


def timeformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 10:15:00", "%Y-%m-%d %H:%M:%S")
    Example output: '11:15:00'

    "timeformat" is used in summary tables where space is tight and times are shown on their own line.
    Currently (Jan 2018) only used in admin app tables.
    """
    return _format_date(value, default_value, DISPLAY_TIME_FORMAT, localize=False)


def displaytimeformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 08:15:00", "%Y-%m-%d %H:%M:%S")
    Example output: '9:15am BST'

    "timeformat_with_tz" is used for events which are less than a few hours away.
    Currently (Mar 2019) only used in timeout warnings to users.
    """
    return _format_date(value, default_value, DISPLAY_TIME_TZ_FORMAT, localize=True)


def shortdateformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 10:15:00", "%Y-%m-%d %H:%M:%S")
    Example output: '25 July'

    "shortdateformat" was designed for use in summary tables where space is tight and dates are shown on their own line.
    The original intended use was in conjunction with "timeformat" in admin app summary tables.
    It is now (Jan 2018) also used in briefs-frontend on the "Publish your requirements" page only.

    ** USE OUR STANDARD dateformat RATHER THAN THIS UNLESS THERE IS A GOOD REASON NOT TO **
    """
    return _format_date(value, default_value, DISPLAY_SHORT_DATE_FORMAT, localize=False)


def nodaydateformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 10:15:00", "%Y-%m-%d %H:%M:%S")
    Example output: '25 July 2018'

    `nodaydateformat` is used *only* when generating framework agreement signature pages. If you're thinking about
    using this, you probably want to use `dateformat` instead.
    """
    return _format_date(value, default_value, DISPLAY_NO_DAY_DATE_FORMAT, localize=False)


def dateformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 10:15:00", "%Y-%m-%d %H:%M:%S")
    Example output: 'Wednesday 25 July 2018'

    "dateformat" is the standard format used for dates in our frontend apps.
    If you're displaying a date on a page then this is probably the one you want.
    Examples of use:
      * dates that opportunities are created/published/closing/closed/withdrawn
      * dates that framework communication files were published
      * dates of revisions made to services (shown to admin users)
    """
    return _format_date(value, default_value, DISPLAY_DATE_FORMAT, localize=False)


def monthyearformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 10:15:00", "%Y-%m-%d %H:%M:%S") OR "2010-01-01T13:00:00.000000Z"
    Example output: 'July 2018'

    This is used when precision is not required and we are trying to remind a user roughly when a historical event
    occurred. The only place this is currently used is when offering a supplier the ability to reuse their declaration
    from a previous framework and we want to remind them roughly when those answers were provided.
    """
    return _format_date(value, default_value, DISPLAY_MONTH_YEAR_FORMAT, localize=False)


def datetimeformat(value, default_value=None, localize=True):
    """
    Example value: datetime.strptime("2018-07-25 23:59:59", "%Y-%m-%d %H:%M:%S")
    Example output: 'Thursday 26 July 2018 at 12:59am BST'
    Example output (localize=False): 'Wednesday 25 July 2018 at 11:59pm UTC'

    "datetimeformat" is the standard format used for timestamps (i.e. date and time) in our frontend apps.
    If you're displaying a date and time on a page then this is probably the one you want (unless it is a
    brief closing date, in which case use "utcdatetimeformat" below).
    Examples of use:
      * dates that opportunities are created/published/closing/closed/withdrawn
      * dates that framework communication files were published
      * dates of revisions made to services (shown to admin users)
    """
    return _format_date(value, default_value, DISPLAY_DATETIME_FORMAT, localize=localize)


def iso_datetime_format(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 23:59:59", "%Y-%m-%d %H:%M:%S")
    Example output: '2018-07-25T23:59:59.000000Z'
    :param value: some datetime that you want to format using ISO 8601 formatting
    :param default_value: default to return if value is None
    :return: string formatted per ISO 8601 in UTC, assuming the datetime passed in was UTC
    """
    return _format_date(value, default_value, DATETIME_FORMAT, localize=False)


def utctoshorttimelongdateformat(value, default_value=None, localize=True):
    """
    Example value: datetime.strptime("2010-01-01 13:00:00", "%Y-%m-%d %H:%M:%S") OR "2010-01-01T13:00:00.000000Z"
    Example output: '1pm GMT, Friday 1 January 2010'

    "utctoshorttimelongdateformat" takes a datetime object or ISO8601 datetime string, localizes it to GMT/BST (as
    appropriate), and then outputs a string with a short time description followed by a long date description including
    the day of the week.

    By default, this will localize the output string so that it shows GMT/BST as appropriate for the target datetime.

    This is primarily used for important framework lifecycle events during the application process.
    """
    formatted_datetime = _format_date(value, default_value, DISPLAY_SHORTTIME_LONGDATE_FORMAT, localize=localize)
    return formatted_datetime


def utcdatetimeformat(value, default_value=None):
    """
    Example value: datetime.strptime("2018-07-25 23:59:59", "%Y-%m-%d %H:%M:%S")
    Example output: 'Wednesday 25 July 2018 at 11:59pm GMT'

    "utcdatetimeformat" forces the timestamp to be output to be in GMT, no matter what the local time is.

    This is intended *ONLY* to be used for *DEADLINES* that are stored in the database as 23:59:59 (UTC+00).
    In this case localisation would display the date as 12.59am on the next day (due to daylight saving),
    potentially causing confusion for buyers/suppliers.
    This is currently (Jan 2018) used wherever we show the closing date for briefs or brief clarification questions.
    """
    local_format = datetimeformat(value, default_value)
    if local_format:
        if "11:59pm" not in local_format:
            # Force UTC+00 if the date has rolled over to the next day and append timezone
            return _use_gmt_timezone(datetimeformat(value, default_value, localize=False))
        return _use_gmt_timezone(local_format)
    return None


def datetodatetimeformat(value):
    """
    Example value: "2018-07-25"
    Example output: 'Wednesday 25 July 2018'

    "datetodatetimeformat" is a convenience function to convert a date *STRING* from one format to another.
    This is currently (Jan 2018) only used to reformat the "startDate" from briefs, which is submitted by
    buyers in DATE_FORMAT but that we want to show to other users of the site in our standard DISPLAY_DATE_FORMAT.

    :param value: a STRING that matches our DATE_FORMAT (eg "2018-01-25")
    :return: a STRING for the same DATE that matches our DISPLAY_DATE_FORMAT (eg "Thursday 25 January 2018")
    """
    try:
        date = datetime.strptime(value, DATE_FORMAT)
        return dateformat(date)
    except ValueError:
        return value


EUROPE_LONDON = pytz.timezone("Europe/London")


def _use_gmt_timezone(date_string):
    return re.sub(r"UTC|BST", "GMT", date_string)


def get_localized_datetime(value: Union[str, datetime], localize: bool = True) -> datetime:
    """
    Given a datetime or an ISO timestamp, will return a London-timezone-localized datetime
    """
    if not isinstance(value, datetime):
        value = datetime.strptime(value, DATETIME_FORMAT)

    if value.tzinfo is None:
        # assume naive datetimes are UTC
        value = pytz.utc.localize(value)

    if localize:
        value = value.astimezone(EUROPE_LONDON)

    return value


def _format_date(value, default_value, fmt, localize=True):
    if not value:
        return default_value

    value = get_localized_datetime(value, localize=localize).strftime(fmt)

    # en_GB locale uses uppercase AM/PM which contravenes our style guide
    value = value.replace('AM', 'am').replace('PM', 'pm').replace(" 0", " ")

    return value



from datetime import timedelta

from flask import current_app, request


DEFAULT_DM_COOKIE_PROBE_COOKIE_NAME = "dm_cookie_probe"
DEFAULT_DM_COOKIE_PROBE_COOKIE_VALUE = "yum"
DEFAULT_DM_COOKIE_PROBE_COOKIE_MAX_AGE = timedelta(days=365).total_seconds()
DEFAULT_DM_COOKIE_PROBE_EXPECT_PRESENT = False


def init_app(app):
    """
        Unconditionally sets a long-lived cookie on all responses from `app`, which we will
        easily be able to check for when trying to determine whether cookies are working for a
        particular client at all.
    """
    app.config.setdefault("DM_COOKIE_PROBE_COOKIE_NAME", DEFAULT_DM_COOKIE_PROBE_COOKIE_NAME)
    app.config.setdefault("DM_COOKIE_PROBE_COOKIE_VALUE", DEFAULT_DM_COOKIE_PROBE_COOKIE_VALUE)
    app.config.setdefault("DM_COOKIE_PROBE_COOKIE_MAX_AGE", DEFAULT_DM_COOKIE_PROBE_COOKIE_MAX_AGE)
    app.config.setdefault("DM_COOKIE_PROBE_COOKIE_EXPECT_PRESENT", DEFAULT_DM_COOKIE_PROBE_EXPECT_PRESENT)

    @app.after_request
    def set_probe_cookie(response):
        response.set_cookie(
            app.config["DM_COOKIE_PROBE_COOKIE_NAME"],
            app.config["DM_COOKIE_PROBE_COOKIE_VALUE"],
            max_age=app.config["DM_COOKIE_PROBE_COOKIE_MAX_AGE"],
        )
        return response


def expected_probe_cookie_missing() -> bool:
    # DM_COOKIE_PROBE_EXPECT_PRESENT controls whether we expect the probe cookie to be present at all - allowing the
    # check to be rolled out and enabled gracefully among multiple frontends.
    is_missing = current_app.config.get(
        "DM_COOKIE_PROBE_EXPECT_PRESENT",
        DEFAULT_DM_COOKIE_PROBE_EXPECT_PRESENT
    ) and (
        request.cookies.get(
            current_app.config.get("DM_COOKIE_PROBE_COOKIE_NAME", DEFAULT_DM_COOKIE_PROBE_COOKIE_NAME)
        ) != current_app.config.get("DM_COOKIE_PROBE_COOKIE_VALUE", DEFAULT_DM_COOKIE_PROBE_COOKIE_VALUE)
    )

    if is_missing:
        current_app.logger.info(
            "cookie_probe.failed: cookies probably not working for this user"
        )

    return bool(is_missing)



import os


def get_api_endpoint_from_stage(stage, app='api'):
    """Return the full URL of given API or Search API environment.

    :param stage: environment name. Can be one of 'preview', 'staging',
                  'production' or 'dev' (aliases: 'local', 'development').
    :param app: should be either 'api' or 'search-api'

    """

    stage_domains = {
        'preview': 'https://{}.preview.marketplace.team'.format(app),
        'staging': 'https://{}.staging.marketplace.team'.format(app),
        'production': 'https://{}.digitalmarketplace.service.gov.uk'.format(app),
        'nft': 'https://{}.nft.marketplace.team'.format(app),
        'uat': 'https://{}.uat.marketplace.team'.format(app),
    }

    dev_ports = {
        "api": os.getenv("DM_API_PORT", 5000),
        "search-api": os.getenv("DM_SEARCH_API_PORT", 5009),
        "antivirus-api": os.getenv("DM_ANTIVIRUS_API_PORT", 5008),
    }

    if stage in ['local', 'dev', 'development']:
        return 'http://localhost:{}'.format(dev_ports[app])

    return stage_domains[stage]


def get_web_url_from_stage(stage):
    """Return the full URL of given web environment.

    :param stage: environment name. Can be one of 'preview', 'staging',
                  'production' or 'dev' (aliases: 'local', 'development').
    """
    if stage in ['local', 'dev', 'development']:
        return 'http://localhost'

    stage_domains = {
        'preview': 'https://www.preview.marketplace.team',
        'staging': 'https://www.staging.marketplace.team',
        'production': 'https://www.digitalmarketplace.service.gov.uk',
        'nft': 'https://www.nft.marketplace.team',
        'uat': 'https://www.uat.marketplace.team',
    }
    return stage_domains[stage]


def get_assets_endpoint_from_stage(stage):
    if stage in ['local', 'dev', 'development']:
        # Static files are not served via nginx for local environments
        raise NotImplementedError()

    return get_api_endpoint_from_stage(stage, 'assets')



from werkzeug.middleware.proxy_fix import ProxyFix


class CustomProxyFix(object):
    def __init__(self, app, forwarded_proto):
        self.app = ProxyFix(app, x_for=1, x_proto=1, x_host=1)
        self.forwarded_proto = forwarded_proto

    def __call__(self, environ, start_response):
        environ.update({
            "HTTP_X_FORWARDED_PROTO": self.forwarded_proto
        })
        return self.app(environ, start_response)


def init_app(app):
    app.wsgi_app = CustomProxyFix(app.wsgi_app,
                                  app.config.get('DM_HTTP_PROTO', 'http'))



class Attribute(object):
    """Wrapper to handle accessing an attribute in service_data"""

    def __init__(
        self,
        value,
        question_type,
        label='',
        optional=False,
    ):
        self.label = label
        self.answer_required = False
        if value in ['', [], None]:
            self.value = ''
            self.type = 'text'
            self.assurance = False
            if not optional:
                self.answer_required = True
        else:
            self.value, self.assurance = self._unpack_assurance(value)
            self.type = question_type

    def _unpack_assurance(self, value):
        if (
            isinstance(value, dict) and
            'assurance' in value
        ):
            if (value['assurance'] == 'Service provider assertion'):
                assurance = False
            else:
                assurance = lowercase_first_character_unless_part_of_acronym(
                    value['assurance']
                )
            return (value['value'], assurance)
        else:
            return (value, False)


def lowercase_first_character_unless_part_of_acronym(string):
    if not string:
        return ''
    if string[1:2] == string[1:2].upper():
        return string
    return string[:1].lower() + string[1:]



import hashlib
import codecs


class AssetFingerprinter():
    """
        Get a unique hash for an asset file, so that it doesn't stay cached
        when it changes

        Usage:

            Config.py:
            base_template_data.asset_fingerprinter = AssetFingerprinter(
                asset_root='/suppliers/static/'
            )

            _base_template.html:
            {{ asset_fingerprinter.get_url('stylesheets/application.css') }}

        * 'app/static' is assumed to be the root for all asset files
    """

    def __init__(self, asset_root='/static/', filesystem_path='app/static/'):
        self._cache = {}
        self._asset_root = asset_root
        self._filesystem_path = filesystem_path

    def get_url(self, asset_path):
        if asset_path not in self._cache:
            self._cache[asset_path] = (
                self._asset_root +
                asset_path +
                '?' +
                self.get_asset_fingerprint(self._filesystem_path + asset_path)
            )
        return self._cache[asset_path]

    def get_asset_fingerprint(self, asset_file_path):
        return hashlib.md5(
            self.get_asset_file_contents(asset_file_path).encode('utf-8')
        ).hexdigest()

    def get_asset_file_contents(self, asset_file_path):
        with codecs.open(asset_file_path, encoding='utf-8') as asset_file:
            contents = asset_file.read()
        return contents



from functools import partial
import warnings

from flask import render_template, render_template_string
from flask_gzip import Gzip

from dmutils.timing import logged_duration


SLOW_RENDER_THRESHOLD = 0.25


_logged_duration_partial = partial(
    logged_duration,
    condition=lambda log_context: (
        logged_duration.default_condition(log_context)  # type: ignore
        or log_context["duration_real"] > SLOW_RENDER_THRESHOLD
    ),
)  # type: ignore


timed_render_template = _logged_duration_partial(
    message="Spent {duration_real}s in render_template",
)(render_template)
timed_render_template.__doc__ = """
    This is a simple ``logged_duration``-wrapped version of flask's ``render_template`` which will output a ``DEBUG``
    log message either when the request has been sent with zipkin "sample" mode or the request takes more than
    ``SLOW_RENDER_THRESHOLD`` seconds. The idea being in frontends we can freely use this in place of flask's regular
    ``render_template`` and get some insight as to the behaviour of the view "for free".
"""

timed_render_template_string = _logged_duration_partial(
    message="Spent {duration_real}s in render_template_string",
)(render_template_string)
timed_render_template_string.__doc__ = """
    See ``timed_render_template``, only for ``render_template_string``.
"""


class DMGzipMiddleware(Gzip):
    compress_by_default: bool

    def __init__(self, *args, compress_by_default: bool = False, **kwargs):
        kwargs.setdefault("minimum_size", 8192)
        self.compress_by_default = compress_by_default
        super().__init__(*args, **kwargs)

    def after_request(self, response):
        x_compression_safe = response.headers.pop("X-Compression-Safe", None)
        known_values = {"0": False, "1": True}
        compress = known_values.get(x_compression_safe, self.compress_by_default)

        if x_compression_safe not in known_values and x_compression_safe != "" and x_compression_safe is not None:
            warnings.warn(
                f"{self.__class__.__name__} received unknown X-Compression-Safe header value {x_compression_safe!r} - "
                "falling back to default"
            )

        # flask_gzip makes the minimum_size comparison itself, but we want to avoid outputting a misleading
        # logged_duration message if it's going to be prevented in the superclass.
        if compress and len(response.get_data()) >= self.minimum_size:
            with logged_duration(message="Spent {duration_real}s compressing response"):
                return super().after_request(response)
        else:
            return response
